class MSGTYPE(object):
    CMD = 1
    OBJ = 2
    OBJ_REQ = 3
    OBJ_DEL = 4
    EXCEPTION = 5
    IS_NONE = 6
    GET_SHAPE = 7
    SEARCH = 8
    FORCE_OBJ_DEL = 9


# Build automatically the reverse map from codes to messaging types
code2MSGTYPE = {}
for code in dir(MSGTYPE):
    if "__" not in code:
        key = getattr(MSGTYPE, code)
        code2MSGTYPE[key] = code



"""
This file tests the ability for serde.py to convert complex types into
simple python types which are serializable by standard serialization tools.
For more on how/why this works, see serde.py directly.
"""
from syft.serde import native_serde
from syft.serde import serde
from syft.serde import torch_serde

import syft
from syft.exceptions import CompressionNotFoundException
from syft.frameworks.torch import pointers

import msgpack
import numpy
import pytest
import torch
from torch import Tensor


def test_tuple_simplify():
    """This tests our ability to simplify tuple types.

    This test is pretty simple since tuples just serialize to
    themselves, with a tuple wrapper with the correct ID (1)
    for tuples so that the detailer knows how to interpret it."""

    input = ("hello", "world")
    tuple_detail_index = serde.detailers.index(native_serde._detail_collection_tuple)
    str_detail_index = serde.detailers.index(native_serde._detail_str)
    target = (
        tuple_detail_index,
        ((str_detail_index, (b"hello",)), (str_detail_index, (b"world",))),
    )
    assert serde._simplify(input) == target


def test_list_simplify():
    """This tests our ability to simplify list types.

    This test is pretty simple since lists just serialize to
    themselves, with a tuple wrapper with the correct ID (2)
    for lists so that the detailer knows how to interpret it."""

    input = ["hello", "world"]
    list_detail_index = serde.detailers.index(native_serde._detail_collection_list)
    str_detail_index = serde.detailers.index(native_serde._detail_str)
    target = (list_detail_index, ((str_detail_index, (b"hello",)), (str_detail_index, (b"world",))))
    assert serde._simplify(input) == target


def test_set_simplify():
    """This tests our ability to simplify set objects.

    This test is pretty simple since sets just serialize to
    lists, with a tuple wrapper with the correct ID (3)
    for sets so that the detailer knows how to interpret it."""

    input = set(["hello", "world"])
    set_detail_index = serde.detailers.index(native_serde._detail_collection_set)
    str_detail_index = serde.detailers.index(native_serde._detail_str)
    target = (set_detail_index, [(str_detail_index, (b"hello",)), (str_detail_index, (b"world",))])
    assert serde._simplify(input)[0] == target[0]
    assert set(serde._simplify(input)[1]) == set(target[1])


def test_float_simplify():
    """This tests our ability to simplify float objects.

    This test is pretty simple since floats just serialize to
    themselves, with no tuple/id necessary."""

    input = 5.6
    target = 5.6
    assert serde._simplify(input) == target


def test_int_simplify():
    """This tests our ability to simplify int objects.

    This test is pretty simple since ints just serialize to
    themselves, with no tuple/id necessary."""

    input = 5
    target = 5
    assert serde._simplify(input) == target


def test_string_simplify():
    """This tests our ability to simplify string objects.

    This test is pretty simple since strings just serialize to
    themselves, with no tuple/id necessary."""

    input = "hello"
    target = (serde.detailers.index(native_serde._detail_str), (b"hello",))
    assert serde._simplify(input) == target


def test_dict_simplify():
    """This tests our ability to simplify dict objects.

    This test is pretty simple since dicts just serialize to
    themselves, with a tuple wrapper with the correct ID
    for dicts so that the detailer knows how to interpret it."""

    input = {"hello": "world"}
    detail_dict_index = serde.detailers.index(native_serde._detail_dictionary)
    detail_str_index = serde.detailers.index(native_serde._detail_str)
    target = (
        detail_dict_index,
        [((detail_str_index, (b"hello",)), (detail_str_index, (b"world",)))],
    )
    assert serde._simplify(input) == target


def test_range_simplify():
    """This tests our ability to simplify range objects.

    This test is pretty simple since range objs just serialize to
    themselves, with a tuple wrapper with the correct ID (5)
    for dicts so that the detailer knows how to interpret it."""

    input = range(1, 3, 4)
    target = (serde.detailers.index(native_serde._detail_range), (1, 3, 4))
    assert serde._simplify(input) == target


def test_torch_tensor_simplify():
    """This tests our ability to simplify torch.Tensor objects

    At the time of writing, tensors simplify to a tuple where the
    first value in the tuple is the tensor's ID and the second
    value is a serialized version of the Tensor (serialized
    by PyTorch's torch.save method)
    """

    # create a tensor
    input = Tensor(numpy.random.random((100, 100)))

    # simplify the tnesor
    output = serde._simplify(input)

    # make sure outer type is correct
    assert type(output) == tuple

    # make sure the object type ID is correct
    # (0 for torch.Tensor)
    assert serde.detailers[output[0]] == torch_serde._detail_torch_tensor

    # make sure inner type is correct
    assert type(output[1]) == tuple

    # make sure ID is correctly encoded
    assert output[1][0] == input.id

    # make sure tensor data type is correct
    assert type(output[1][1]) == bytes


def test_ndarray_simplify():
    """This tests our ability to simplify numpy.array objects

    At the time of writing, arrays simplify to an object inside
    of a tuple which specifies the ID for the np.array type (6) so
    that the detailer knows to turn the simplifed form to a np.array
    """

    input = numpy.random.random((100, 100))
    output = serde._simplify(input)

    # make sure simplified type ID is correct
    assert serde.detailers[output[0]] == torch_serde._detail_ndarray

    # make sure serialized form is correct
    assert type(output[1][0]) == bytes
    assert output[1][1] == input.shape
    assert output[1][2] == input.dtype.name


def test_ellipsis_simplify():
    """Make sure ellipsis simplifies correctly."""
    assert serde.detailers[serde._simplify(Ellipsis)[0]] == native_serde._detail_ellipsis

    # the simplified ellipsis (empty object)
    assert serde._simplify(Ellipsis)[1] == b""


def test_torch_device_simplify():
    """Test the simplification of torch.device"""
    device = torch.device("cpu")

    assert serde.detailers[serde._simplify(device)[0]] == torch_serde._detail_torch_device

    # the simplified torch.device
    assert serde._simplify(device)[1] == "cpu"


def test_pointer_tensor_simplify():
    """Test the simplification of PointerTensor"""

    alice = syft.VirtualWorker(syft.torch.hook, id="alice")
    input_tensor = pointers.PointerTensor(id=1000, location=alice, owner=alice)

    output = serde._simplify(input_tensor)

    assert output[1][0] == input_tensor.id
    assert output[1][1] == input_tensor.id_at_location
    assert output[1][2] == input_tensor.owner.id


@pytest.mark.parametrize("compress", [True, False])
def test_torch_Tensor(compress):
    if compress:
        syft.serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        syft.serde._apply_compress_scheme = serde.apply_no_compression

    t = Tensor(numpy.random.random((100, 100)))
    t_serialized = serde.serialize(t)
    t_serialized_deserialized = serde.deserialize(t_serialized)
    assert (t == t_serialized_deserialized).all()


@pytest.mark.parametrize("compress", [True, False])
def test_torch_Tensor_convenience(compress):
    """This test evaluates torch.Tensor.serialize()

    As opposed to using syft.serde.serialize(), torch objects
    have a convenience function which lets you call .serialize()
    directly on the tensor itself. This tests to makes sure it
    works correctly."""
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    t = Tensor(numpy.random.random((100, 100)))
    t_serialized = t.serialize()
    t_serialized_deserialized = serde.deserialize(t_serialized)
    assert (t == t_serialized_deserialized).all()


@pytest.mark.parametrize("compress", [True, False])
def test_tuple(compress):
    # Test with a simple datatype
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    tuple = (1, 2)
    tuple_serialized = serde.serialize(tuple)
    tuple_serialized_deserialized = serde.deserialize(tuple_serialized)
    assert tuple == tuple_serialized_deserialized

    # Test with a complex data structure
    tensor_one = Tensor(numpy.random.random((100, 100)))
    tensor_two = Tensor(numpy.random.random((100, 100)))
    tuple = (tensor_one, tensor_two)
    tuple_serialized = serde.serialize(tuple)
    tuple_serialized_deserialized = serde.deserialize(tuple_serialized)
    # `assert tuple_serialized_deserialized == tuple` does not work, therefore it's split
    # into 3 assertions
    assert type(tuple_serialized_deserialized) == type(tuple)
    assert (tuple_serialized_deserialized[0] == tensor_one).all()
    assert (tuple_serialized_deserialized[1] == tensor_two).all()


@pytest.mark.parametrize("compress", [True, False])
def test_bytearray(compress):
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    bytearr = bytearray("This is a teststring", "utf-8")
    bytearr_serialized = serde.serialize(bytearr)
    bytearr_serialized_desirialized = serde.deserialize(bytearr_serialized)
    assert bytearr == bytearr_serialized_desirialized

    bytearr = bytearray(numpy.random.random((100, 100)))
    bytearr_serialized = serde.serialize(bytearr)
    bytearr_serialized_desirialized = serde.deserialize(bytearr_serialized)
    assert bytearr == bytearr_serialized_desirialized


@pytest.mark.parametrize("compress", [True, False])
def test_ndarray_serde(compress):
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression
    arr = numpy.random.random((100, 100))
    arr_serialized = serde.serialize(arr)

    arr_serialized_deserialized = serde.deserialize(arr_serialized)

    assert numpy.array_equal(arr, arr_serialized_deserialized)


@pytest.mark.parametrize("compress_scheme", [serde.LZ4, serde.ZSTD, serde.NO_COMPRESSION])
def test_compress_decompress(compress_scheme):
    if compress_scheme == serde.LZ4:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    elif compress_scheme == serde.ZSTD:
        serde._apply_compress_scheme = serde.apply_zstd_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    original = msgpack.dumps([1, 2, 3])
    compressed = serde._compress(original)
    decompressed = serde._decompress(compressed)
    assert type(compressed) == bytes
    assert original == decompressed


@pytest.mark.parametrize("compress_scheme", [serde.LZ4, serde.ZSTD, serde.NO_COMPRESSION])
def test_compressed_serde(compress_scheme):
    if compress_scheme == serde.LZ4:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    elif compress_scheme == serde.ZSTD:
        serde._apply_compress_scheme = serde.apply_zstd_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    # using numpy.ones because numpy.random.random is not compressed.
    arr = numpy.ones((100, 100))

    arr_serialized = serde.serialize(arr)

    arr_serialized_deserialized = serde.deserialize(arr_serialized)
    assert numpy.array_equal(arr, arr_serialized_deserialized)


@pytest.mark.parametrize("compress", [True, False])
def test_dict(compress):
    # Test with integers
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression
    _dict = {1: 1, 2: 2, 3: 3}
    dict_serialized = serde.serialize(_dict)
    dict_serialized_deserialized = serde.deserialize(dict_serialized)
    assert _dict == dict_serialized_deserialized

    # Test with strings
    _dict = {"one": 1, "two": 2, "three": 3}
    dict_serialized = serde.serialize(_dict)
    dict_serialized_deserialized = serde.deserialize(dict_serialized)
    assert _dict == dict_serialized_deserialized

    # Test with a complex data structure
    tensor_one = Tensor(numpy.random.random((100, 100)))
    tensor_two = Tensor(numpy.random.random((100, 100)))
    _dict = {0: tensor_one, 1: tensor_two}
    dict_serialized = serde.serialize(_dict)
    dict_serialized_deserialized = serde.deserialize(dict_serialized)
    # `assert dict_serialized_deserialized == _dict` does not work, therefore it's split
    # into 3 assertions
    assert type(dict_serialized_deserialized) == type(_dict)
    assert (dict_serialized_deserialized[0] == tensor_one).all()
    assert (dict_serialized_deserialized[1] == tensor_two).all()


@pytest.mark.parametrize("compress", [True, False])
def test_range_serde(compress):
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    _range = range(1, 2, 3)

    range_serialized = serde.serialize(_range)
    range_serialized_deserialized = serde.deserialize(range_serialized)

    assert _range == range_serialized_deserialized


@pytest.mark.parametrize("compress", [True, False])
def test_list(compress):
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    # Test with integers
    _list = [1, 2]
    list_serialized = serde.serialize(_list)
    list_serialized_deserialized = serde.deserialize(list_serialized)
    assert _list == list_serialized_deserialized

    # Test with strings
    _list = ["hello", "world"]
    list_serialized = serde.serialize(_list)
    list_serialized_deserialized = serde.deserialize(list_serialized)
    assert _list == list_serialized_deserialized

    # Test with a complex data structure
    tensor_one = Tensor(numpy.ones((100, 100)))
    tensor_two = Tensor(numpy.ones((100, 100)) * 2)
    _list = (tensor_one, tensor_two)

    list_serialized = serde.serialize(_list)
    if compress:
        assert list_serialized[0] == serde.LZ4
    else:
        assert list_serialized[0] == serde.NO_COMPRESSION

    list_serialized_deserialized = serde.deserialize(list_serialized)
    # `assert list_serialized_deserialized == _list` does not work, therefore it's split
    # into 3 assertions
    assert type(list_serialized_deserialized) == type(_list)
    assert (list_serialized_deserialized[0] == tensor_one).all()
    assert (list_serialized_deserialized[1] == tensor_two).all()


@pytest.mark.parametrize("compress", [True, False])
def test_set(compress):
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    # Test with integers
    _set = set([1, 2])
    set_serialized = serde.serialize(_set)

    set_serialized_deserialized = serde.deserialize(set_serialized)
    assert _set == set_serialized_deserialized

    # Test with strings
    _set = set(["hello", "world"])
    set_serialized = serde.serialize(_set)
    set_serialized_deserialized = serde.deserialize(set_serialized)
    assert _set == set_serialized_deserialized

    # Test with a complex data structure
    tensor_one = Tensor(numpy.ones((100, 100)))
    tensor_two = Tensor(numpy.ones((100, 100)) * 2)
    _set = (tensor_one, tensor_two)

    set_serialized = serde.serialize(_set)
    if compress:
        assert set_serialized[0] == serde.LZ4
    else:
        assert set_serialized[0] == serde.NO_COMPRESSION

    set_serialized_deserialized = serde.deserialize(set_serialized)
    # `assert set_serialized_deserialized == _set` does not work, therefore it's split
    # into 3 assertions
    assert type(set_serialized_deserialized) == type(_set)
    assert (set_serialized_deserialized[0] == tensor_one).all()
    assert (set_serialized_deserialized[1] == tensor_two).all()


@pytest.mark.parametrize("compress", [True, False])
def test_slice(compress):
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    s = slice(0, 100, 2)
    x = numpy.random.rand(100)
    s_serialized = serde.serialize(s)
    s_serialized_deserialized = serde.deserialize(s_serialized)

    assert type(s) == type(s_serialized_deserialized)
    assert (x[s] == x[s_serialized_deserialized]).all()

    s = slice(40, 50)
    x = numpy.random.rand(100)
    s_serialized = serde.serialize(s)
    s_serialized_deserialized = serde.deserialize(s_serialized)

    assert type(s) == type(s_serialized_deserialized)
    assert (x[s] == x[s_serialized_deserialized]).all()


@pytest.mark.parametrize("compress", [True, False])
def test_float(compress):
    if compress:
        serde._apply_compress_scheme = serde.apply_lz4_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    x = 0.5
    y = 1.5

    x_serialized = serde.serialize(x)
    x_serialized_deserialized = serde.deserialize(x_serialized)

    y_serialized = serde.serialize(y)
    y_serialized_deserialized = serde.deserialize(y_serialized)

    assert x_serialized_deserialized == x
    assert y_serialized_deserialized == y


@pytest.mark.parametrize(
    "compress, compress_scheme",
    [
        (True, serde.LZ4),
        (False, serde.LZ4),
        (True, serde.ZSTD),
        (False, serde.ZSTD),
        (True, serde.NO_COMPRESSION),
        (False, serde.NO_COMPRESSION),
    ],
)
def test_hooked_tensor(compress, compress_scheme):
    if compress:
        if compress_scheme == serde.LZ4:
            serde._apply_compress_scheme = serde.apply_lz4_compression
        elif compress_scheme == serde.ZSTD:
            serde._apply_compress_scheme = serde.apply_zstd_compression
        else:
            serde._apply_compress_scheme = serde.apply_no_compression
    else:
        serde._apply_compress_scheme = serde.apply_no_compression

    t = Tensor(numpy.ones((100, 100)))
    t_serialized = serde.serialize(t)
    assert (
        t_serialized[0] == compress_scheme if compress else t_serialized[0] == serde.NO_COMPRESSION
    )
    t_serialized_deserialized = serde.deserialize(t_serialized)
    assert (t == t_serialized_deserialized).all()


def test_pointer_tensor(hook, workers):
    serde._apply_compress_scheme = serde.apply_no_compression
    t = pointers.PointerTensor(
        id=1000, location=workers["alice"], owner=workers["alice"], id_at_location=12345
    )
    t_serialized = serde.serialize(t)
    t_serialized_deserialized = serde.deserialize(t_serialized)
    assert t.id == t_serialized_deserialized.id
    assert t.location.id == t_serialized_deserialized.location.id
    assert t.id_at_location == t_serialized_deserialized.id_at_location


@pytest.mark.parametrize("id", [1000, "1000"])
def test_pointer_tensor_detail(id):
    alice = syft.VirtualWorker(syft.torch.hook, id=id)
    x = torch.tensor([1, -1, 3, 4])
    x_ptr = x.send(alice)
    x_ptr = 2 * x_ptr
    x_back = x_ptr.get()
    assert (x_back == 2 * x).all()


def test_numpy_tensor_serde():
    serde._apply_compress_scheme = serde.apply_lz4_compression

    serde._serialize_tensor = syft.serde.numpy_tensor_serializer
    serde._deserialize_tensor = syft.serde.numpy_tensor_deserializer

    tensor = torch.tensor(numpy.ones((10, 10)), requires_grad=False)

    tensor_serialized = serde.serialize(tensor)
    assert tensor_serialized[0] != serde.NO_COMPRESSION
    tensor_deserialized = serde.deserialize(tensor_serialized)

    # Back to Pytorch serializer
    serde._serialize_tensor = syft.serde.torch_tensor_serializer
    serde._deserialize_tensor = syft.serde.torch_tensor_deserializer

    assert torch.eq(tensor_deserialized, tensor).all()


@pytest.mark.parametrize("compress", [True, False])
def test_additive_sharing_tensor_serde(compress, workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x = torch.tensor([[3.1, 4.3]]).fix_prec().share(alice, bob, crypto_provider=james)

    additive_sharing_tensor = x.child.child
    data = syft.AdditiveSharingTensor.simplify(additive_sharing_tensor)
    additive_sharing_tensor_reconstructed = syft.AdditiveSharingTensor.detail(
        syft.hook.local_worker, data
    )

    assert additive_sharing_tensor_reconstructed.field == additive_sharing_tensor.field

    assert (
        additive_sharing_tensor_reconstructed.child.keys() == additive_sharing_tensor.child.keys()
    )


@pytest.mark.parametrize("compress", [True, False])
def test_fixed_precision_tensor_serde(compress, workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x = (
        torch.tensor([[3.1, 4.3]])
        .fix_prec(base=12, precision_fractional=5)
        .share(alice, bob, crypto_provider=james)
    )

    serialized_x = serde.serialize(x)
    deserialied_x = serde.deserialize(serialized_x)

    assert x.id == deserialied_x.child.id
    assert x.child.field == deserialied_x.child.field
    assert x.child.kappa == deserialied_x.child.kappa
    assert x.child.precision_fractional == deserialied_x.child.precision_fractional
    assert x.child.base == deserialied_x.child.base


def test_serde_object_wrapper_int():
    obj = 4
    obj_wrapper = pointers.ObjectWrapper(obj, id=100)
    msg = serde.serialize(obj_wrapper)

    obj_wrapper_received = serde.deserialize(msg)

    assert obj_wrapper.obj == obj_wrapper_received.obj
    assert obj_wrapper.id == obj_wrapper_received.id


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_serialize_and_deserialize_torch_scriptmodule():  # pragma: no cover
    @torch.jit.script
    def foo(x):
        return x + 2

    bin_message = torch_serde._simplify_script_module(foo)
    foo_loaded = torch_serde._detail_script_module(None, bin_message)

    assert foo.code == foo_loaded.code


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_torch_jit_script_module_serde():  # pragma: no cover
    @torch.jit.script
    def foo(x):
        return x + 2

    msg = serde.serialize(foo)
    foo_received = serde.deserialize(msg)

    assert foo.code == foo_received.code


def test_serde_virtual_worker(hook):
    virtual_worker = syft.VirtualWorker(hook=hook, id="deserialized_worker1")
    # Populate worker
    tensor1, tensor2 = torch.tensor([1.0, 2.0]), torch.tensor([0.0])
    ptr1, ptr2 = tensor1.send(virtual_worker), tensor2.send(virtual_worker)

    serialized_worker = serde.serialize(virtual_worker, force_full_simplification=False)
    deserialized_worker = serde.deserialize(serialized_worker)

    assert virtual_worker.id == deserialized_worker.id


def test_full_serde_virtual_worker(hook):
    virtual_worker = syft.VirtualWorker(hook=hook, id="deserialized_worker2")
    # Populate worker
    tensor1, tensor2 = torch.tensor([1.0, 2.0]), torch.tensor([0.0])
    ptr1, ptr2 = tensor1.send(virtual_worker), tensor2.send(virtual_worker)

    serialized_worker = serde.serialize(virtual_worker, force_full_simplification=True)

    deserialized_worker = serde.deserialize(serialized_worker)

    assert virtual_worker.id == deserialized_worker.id
    assert virtual_worker.auto_add == deserialized_worker.auto_add
    assert len(deserialized_worker._objects) == 2
    assert tensor1.id in deserialized_worker._objects
    assert tensor2.id in deserialized_worker._objects


def test_serde_object_wrapper_traced_module():

    data = torch.tensor([[-1, 2.0], [0, 1.1], [-1, 2.1], [0, 1.2]])

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = torch.nn.Linear(2, 3)

        def forward(self, x):
            x = torch.nn.functional.relu(self.fc1(x))
            return x

    obj = torch.jit.trace(Net(), data)

    obj_wrapper = pointers.ObjectWrapper(obj, id=200)
    msg = serde.serialize(obj_wrapper)

    obj_wrapper_received = serde.deserialize(msg)

    pred_before = obj(data)

    pred_after = obj_wrapper_received.obj(data)

    assert (pred_before == pred_after).all()
    assert obj_wrapper.id == obj_wrapper_received.id



from syft.workers.abstract import AbstractWorker  # noqa: F401
from syft.workers.base import BaseWorker  # noqa: F401
from syft.workers.virtual import VirtualWorker  # noqa: F401

from syft.workers.websocket_client import WebsocketClientWorker  # noqa: F401
from syft.workers.websocket_server import WebsocketServerWorker  # noqa: F401

from syft import dependency_check

__all__ = ["base", "virtual", "websocket_client", "socketio_client", "BaseWorker"]

if dependency_check.keras_available:
    from syft.workers.tfe import TFECluster  # noqa: F401
    from syft.workers.tfe import TFEWorker  # noqa: F401

    __all__.append("tfe")



# This script reads in derivative definitions from derivatives.yaml and writes appropriately
# constructed GradFunc classes to gradients.py. It could probably be optimized.

import re
import yaml

tab = " " * 4


def split_signature(signature):
    return re.match(r"^(\S+)\((.+)\)", signature).groups()


# There is probably a better way to do this
# Need to handle keyword arguments as well
def construct_grad_fn_class(grad_def):
    lines = []

    name, arguments = split_signature(grad_def["name"])

    # This won't work right if keyword arguments are present. I should refactor
    # this with it's own function. TODO
    input_args = arguments.split(", ")
    signature = arguments.replace("self", "self_")

    lines.append(f"class {name.capitalize()}Backward(GradFunc):")
    lines.append(tab + f"def __init__(self, {signature}):")
    lines.append(2 * tab + f"super().__init__(self, {signature})")
    for arg in signature.split(", "):
        lines.append(2 * tab + f"self.{arg} = {arg}")
    lines.append("")

    lines.append(tab + "def gradient(self, grad):")
    for arg in input_args:
        formula = grad_def[arg]
        for in_arg in input_args:
            formula = formula.replace(in_arg, f"self.{in_arg}")
        formula = formula.replace(".self", ".self_").replace("result", "self.result")
        lines.append(2 * tab + f"grad_{arg.replace('self', 'self_')} = {formula}")
    grad_signature = ", ".join([f"grad_{arg}".replace("self", "self_") for arg in input_args])
    lines.append(2 * tab + f"return ({grad_signature},)")

    return "\n".join(lines) + "\n"


if __name__ == "__main__":
    with open("syft/frameworks/torch/tensors/interpreters/derivatives.yaml", "r") as f:
        derivatives = yaml.load(f.read())

    with open("syft/frameworks/torch/tensors/interpreters/gradients.py", "w") as f:
        grad_fn_map = {}
        f.write("# This file is generated from build_gradients.py\n\n")
        f.write("from . gradients_core import *\n\n")
        for definition in derivatives:
            f.write(construct_grad_fn_class(definition))
            f.write("\n")



"""To be extended in the near future."""
from collections import OrderedDict
import logging
import os
import subprocess
import tempfile

import tf_encrypted as tfe


logger = logging.getLogger("tf_encrypted")
_TMP_DIR = tempfile.gettempdir()


class TFEWorker:
    # TODO(Morten) this should be turned into a proxy, with existing code
    # extracted into a new component that's launched via a script

    def __init__(self, host=None, auto_managed=True):
        self.host = host
        self._server_process = None
        self._auto_managed = auto_managed

    def start(self, player_name, cluster):
        """
        Start the worker as a player in the given cluster.
        Depending on whether the worker was constructed with a host or not
        this may launch a subprocess running a TensorFlow server.
        """

        if self.host is None:
            # we're running using a tfe.LocalConfig which doesn't require us to do anything
            return

        config_filename = os.path.join(_TMP_DIR, "tfe.config")
        config = cluster.tfe_config
        config.save(config_filename)

        launch_cmd = "python -m tf_encrypted.player --config {} {}".format(
            config_filename, player_name
        )
        if self._auto_managed:
            self._server_process = subprocess.Popen(launch_cmd.split(" "))
        else:
            logger.info(
                "If not done already, please launch the following "
                "command in a terminal on host %s: '%s'\n"
                "This can be done automatically in a local subprocess by "
                "setting `auto_managed=True` when instantiating a TFEWorker.\n",
                self.host,
                launch_cmd,
            )

    def stop(self):
        """
        Stop the worker. This will shutdown any TensorFlow server launched
        in `start()`.
        """

        if self.host is None:
            # we're running using a tfe.LocalConfig which doesn't require us to do anything
            return

        if self._auto_managed:
            if self._server_process is None:
                return
            self._server_process.kill()
            self._server_process.communicate()
            self._server_process = None
        else:
            logger.info("Please terminate the process on host '%s'.", self.host)

    def connect_to_model(self, input_shape, output_shape, cluster, sess=None):
        """
        Connect to a TF Encrypted model being served by the given cluster.
        
        This must be done before querying the model.
        """

        config = cluster.tfe_config
        tfe.set_config(config)

        prot = tfe.protocol.SecureNN(
            config.get_player("server0"), config.get_player("server1"), config.get_player("server2")
        )
        tfe.set_protocol(prot)

        self._tf_client = tfe.serving.QueueClient(
            input_shape=input_shape, output_shape=output_shape
        )

        if sess is None:
            sess = tfe.Session(config=config)
        self._tf_session = sess

    def query_model(self, data):
        """
        Encrypt data and sent it as input to the model being served.
        
        This will block until a result is ready, and requires that
        a connection to the model has already been established via
        `connect_to_model()`.
        """
        self.query_model_async(data)
        return self.query_model_join()

    def query_model_async(self, data):
        """
        Asynchronous version of `query_model` that will not block until a
        result is ready. Call `query_model_join` to retrive result.
        
        This requires that a connection to the model has already been
        established via `connect_to_model()`.
        """
        self._tf_client.send_input(self._tf_session, data)

    def query_model_join(self):
        """
        Retrives the result from calling `query_model_async`, blocking until
        ready.
        """
        return self._tf_client.receive_output(self._tf_session)


class TFECluster:
    """
    A TFECluster represents a group of TFEWorkers that are aware about each
    other and collectively perform an encrypted computation.
    """

    def __init__(self, *workers):
        tfe_config, player_to_worker_mapping = self._build_cluster(workers)
        self.tfe_config = tfe_config
        self.player_to_worker_mapping = player_to_worker_mapping

    @property
    def workers(self):
        return list(self.player_to_worker_mapping.values())

    def start(self):
        """
        Start all workers in the cluster.
        """
        # Tell the TFE workers to launch TF servers
        for player_name, worker in self.player_to_worker_mapping.items():
            worker.start(player_name, self)

    def stop(self):
        """
        Stop all workers in the cluster.
        """
        for worker in self.workers:
            worker.stop()

    def _build_cluster(self, workers):
        if len(workers) != 3:
            raise ValueError("Expected three workers but {} were given".format(len(workers)))

        player_to_worker_mapping = OrderedDict()
        player_to_worker_mapping["server0"] = workers[0]
        player_to_worker_mapping["server1"] = workers[1]
        player_to_worker_mapping["server2"] = workers[2]

        use_local_config = all(worker.host is None for worker in workers)
        if use_local_config:
            config = tfe.LocalConfig(
                player_names=player_to_worker_mapping.keys(), auto_add_unknown_players=False
            )
            return config, player_to_worker_mapping

        # use tfe.RemoteConfig
        hostmap = OrderedDict(
            [(player_name, worker.host) for player_name, worker in player_to_worker_mapping.items()]
        )
        config = tfe.RemoteConfig(hostmap)

        return config, player_to_worker_mapping



# import torch
# import numpy as np

# from syft.frameworks.torch.tensors.interpreters.polynomial import PolynomialTensor


# def test_sigmoid():
#
#     poly_tensor = PolynomialTensor()
#
#     # x = torch.tensor(np.linspace(-3, 3, 10), dtype=torch.double)
#
#     expected = torch.tensor(
#         [0.0337, 0.0886, 0.1759, 0.2921, 0.4283, 0.5717, 0.7079, 0.8241, 0.9114, 0.9663]
#     )
#
#
#     # allclose function to compare the expected values and approximations with fixed precision
#     # result = poly_tensor.get_val("sigmoid", x)
#
#     # assert torch.abs(expected - expected).max()[0] < 1e-03
#     _ = (expected - expected).max()
#     print(_)
#     assert True

#
# def test_exp():
#
#     poly_tensor = PolynomialTensor()
#
#     x = torch.tensor(np.linspace(-3, 3, 10), dtype=torch.double)
#
#     expected = torch.tensor(
#         [0.0498, 0.0970, 0.1889, 0.3679, 0.7165, 1.1176, 2.9503, 5.1088, 10.2501, 20.2955],
#         dtype=torch.double,
#     )
#
#     # allclose function to compare the expected values and approximations with fixed precision
#     result = poly_tensor.get_val("exp", x)
#     assert torch.allclose(expected, result, atol=1e-03)
#
#
# def test_tanh():
#
#     # Test if tanh approximation works as expected
#
#     poly_tensor = PolynomialTensor()
#
#     x = torch.tensor(np.linspace(-3, 3, 10), dtype=torch.double)
#
#     expected = torch.tensor(
#         [-0.9937, -0.9811, -0.9329, -0.7596, -0.3239, 0.3239, 0.7596, 0.9329, 0.9811, 0.9937],
#         dtype=torch.double,
#     )
#
#     result = poly_tensor.get_val("tanh", x)
#
#     # allclose function to compare the expected values and approximations with fixed precision
#     assert torch.allclose(expected, result, atol=1e-03)
#
#
# def test_interpolate():
#
#     # Test if interpolation function works as expected by verifying an approximation of exponential function
#
#     expected = torch.tensor([1.2220, 2.9582, 7.1763, 20.3064, 54.4606], dtype=torch.double)
#
#     poly_tensor = PolynomialTensor()
#
#     f1 = poly_tensor.interpolate((lambda x: np.exp(x)), np.linspace(0, 10, 50))
#
#     assert torch.allclose(expected, torch.tensor(f1(torch.tensor([0, 1, 2, 3, 4]))), 1e-04)
#
#
# def test_custom_function():
#     poly_tensor = PolynomialTensor()
#     poly_tensor.add_function(
#         "Custom", 10, [[0, 10, 100, 10, poly_tensor.fit_function]], lambda x: x + 2
#     )
#
#     assert round(poly_tensor.get_val("Custom", 3)) == 5
#     assert round(poly_tensor.get_val("Custom", 6)) == 8
#
#
# def test_random_function():
#
#     x = torch.tensor(np.linspace(-3, 3, 10), dtype=torch.double)
#     poly_tensor = PolynomialTensor(function=lambda x: x * 2)
#     scaled_result = poly_tensor.get_val("exp", x.clone())
#
#     poly_tensor = PolynomialTensor()
#     original_result = poly_tensor.get_val("exp", x)
#
#     assert torch.all(torch.eq(scaled_result, torch.mul(original_result, 2)))
#
#
# def test_log_function():
#
#     # Test if log approximation works as expected
#
#     poly_tensor = PolynomialTensor()
#
#     x = torch.tensor(np.linspace(1, 10, 10), dtype=torch.double)
#
#     expected = torch.tensor(
#         [
#             3.7160e-04,
#             6.9319e-01,
#             1.0986e00,
#             1.3863e00,
#             1.6096e00,
#             1.7932e00,
#             1.9526e00,
#             2.1056e00,
#             2.2835e00,
#             2.5537e00,
#         ],
#         dtype=torch.double,
#     )
#
#     result = poly_tensor.get_val("log", x.clone())
#
#     # allclose function to compare the expected values and approximations with fixed precision
#     assert torch.allclose(expected, result, atol=1e-03)
#
#
# def test_exp_taylor():
#
#     expected = torch.tensor(
#         [-0.1076, 0.0664, 0.1852, 0.3677, 0.7165, 1.3956, 2.7180, 5.2867, 10.2325, 19.5933],
#         dtype=torch.double,
#     )
#     poly_tensor = PolynomialTensor()
#     x = torch.tensor(np.linspace(-3, 3, 10), dtype=torch.double)
#     result = poly_tensor.exp(x)
#
#     # allclose function to compare the expected values and approximations with fixed precision
#     assert torch.allclose(expected, result, atol=1e-03)
#
#
# def test_sigmoid_taylor():
#
#     expected = torch.tensor(
#         [0.1000, 0.1706, 0.2473, 0.3392, 0.4447, 0.5553, 0.6608, 0.7527, 0.8294, 0.9000],
#         dtype=torch.double,
#     )
#     poly_tensor = PolynomialTensor()
#     x = torch.tensor(np.linspace(-2, 2, 10), dtype=torch.double)
#     result = poly_tensor.sigmoid(x)
#
#     # allclose function to compare the expected values and approximations with fixed precision
#     assert torch.allclose(expected, result, atol=1e-03)



import torch
from .gradients_core import GradFunc
from .gradients_core import apply_dim_transformations


class AddBackward(GradFunc):
    def __init__(self, self_, other):
        super().__init__(self, self_, other)
        self.self_ = self_
        self.other = other

    def gradient(self, grad):
        grad_self = grad.copy()
        grad_other = grad.copy() if type(self.self_) == type(self.other) else None

        if self.self_.shape != self.other.shape:
            grad_self, grad_other = apply_dim_transformations(
                grad_self, grad_other, self.self_.shape, self.other.shape
            )

        return (grad_self, grad_other)


class SubBackward(GradFunc):
    def __init__(self, self_, other):
        super().__init__(self, self_, other)
        self.self_ = self_
        self.other = other

    def gradient(self, grad):
        grad_self = grad.copy()
        grad_other = grad * -1 if type(self.self_) == type(self.other) else None

        if self.self_.shape != self.other.shape:
            grad_self, grad_other = apply_dim_transformations(
                grad_self, grad_other, self.self_.shape, self.other.shape
            )
        return (grad_self, grad_other)


class SumBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        return ((self.self_ * 0 + 1) * grad,)
        # return (torch.ones(self.self_.shape) * grad, )


class AsinBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        grad_self_ = grad * (-self.self_ * self.self_ + 1).rsqrt()
        return (grad_self_,)


class MulBackward(GradFunc):
    def __init__(self, self_, other):
        super().__init__(self, self_, other)
        self.self_ = self_
        self.other = other

    def gradient(self, grad):
        grad_self_ = grad * self.other
        grad_other = grad * self.self_ if type(self.self_) == type(self.other) else None
        return (grad_self_, grad_other)


class DivBackward(GradFunc):
    def __init__(self, self_, other):
        super().__init__(self, self_, other)
        self.self_ = self_
        self.other = other

    def gradient(self, grad):
        assert isinstance(self.other, int)
        grad_self_ = grad / self.other
        return (grad_self_,)


class PowBackward(GradFunc):
    def __init__(self, self_, power):
        super().__init__(self, self_, power)
        self.self_ = self_
        self.power = power

    def gradient(self, grad):
        power = self.power
        return (power * self.self_ ** (power - 1) * grad,)


class MatmulBackward(GradFunc):
    def __init__(self, self_, other):
        super().__init__(self, self_, other)
        self.self_ = self_
        self.other = other

    def gradient(self, grad):
        grad_self_ = grad @ self.other.t()
        grad_other = self.self_.t() @ grad if type(self.self_) == type(self.other) else None
        return (grad_self_, grad_other)


class TBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        return (grad.t(),)


class SigmoidBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        grad_self_ = grad * self.self_.sigmoid() * (1 - self.self_.sigmoid())
        return (grad_self_,)


class SinBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        grad_self_ = grad * self.self_.cos()
        return (grad_self_,)


class SinhBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        grad_self_ = grad * self.self_.cosh()
        return (grad_self_,)


class SqrtBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        grad_self_ = grad / (2 * self.result)
        return (grad_self_,)


class TanhBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        grad_self_ = grad * (1 - self.self_.tanh() ** 2)
        return (grad_self_,)


class ReluBackward(GradFunc):
    def __init__(self, self_):
        super().__init__(self, self_)
        self.self_ = self_

    def gradient(self, grad):
        zero = self.self_ * 0
        gt_zero = self.self_ > zero
        return (gt_zero * grad,)



import syft


class Module(object):
    pass


class Overloaded:
    def __init__(self):
        self.method = Overloaded.overload_method
        self.function = Overloaded.overload_function
        self.module = Overloaded.overload_module

    @staticmethod
    def overload_method(attr):
        """
        hook args and response for methods that hold the @overloaded.method decorator
        """

        def hook_args(self, *args, **kwargs):
            # Replace all syft tensor with their child attribute
            new_self, new_args, new_kwargs = syft.frameworks.torch.hook_args.unwrap_args_from_method(
                attr.__name__, self, args, kwargs
            )

            # Send it to the appropriate class and get the response
            response = attr(self, new_self, *new_args, **new_kwargs)

            # Put back SyftTensor on the tensors found in the response
            response = syft.frameworks.torch.hook_args.hook_response(
                attr.__name__, response, wrap_type=type(self), wrap_args=self.get_class_attributes()
            )

            return response

        return hook_args

    @staticmethod
    def overload_function(attr):
        """
        hook args and response for functions that hold the @overloaded.function decorator
        """

        def hook_args(*args, **kwargs):

            # TODO have a better way to infer the type of tensor -> this is implies
            # that the first argument is a tensor (even if this is the case > 99%)
            tensor = args[0] if not isinstance(args[0], tuple) else args[0][0]
            cls = type(tensor)

            # Replace all syft tensor with their child attribute
            new_args, new_kwargs, new_type = syft.frameworks.torch.hook_args.unwrap_args_from_function(
                attr.__name__, args, kwargs
            )

            # Send it to the appropriate class and get the response
            response = attr(*new_args, **new_kwargs)

            # Put back SyftTensor on the tensors found in the response
            response = syft.frameworks.torch.hook_args.hook_response(
                attr.__name__, response, wrap_type=cls, wrap_args=tensor.get_class_attributes()
            )

            return response

        return hook_args

    @staticmethod
    def overload_module(attr):

        module = Module()
        attr(module)

        return module


overloaded = Overloaded()



from .abstract import AbstractTensor  # noqa: F401
from .native import TorchTensor  # noqa: F401
from .precision import FixedPrecisionTensor  # noqa: F401
from .additive_shared import AdditiveSharingTensor  # noqa: F401
from .multi_pointer import MultiPointerTensor  # noqa: F401
from .crt_precision import CRTPrecisionTensor  # noqa: F401
from .large_precision import LargePrecisionTensor  # noqa: F401
from .autograd import AutogradTensor  # noqa: F401



from syft.frameworks.torch.pointers.object_pointer import ObjectPointer
from syft.frameworks.torch.pointers.callable_pointer import create_callable_pointer
from syft.frameworks.torch.pointers.callable_pointer import CallablePointer
from syft.frameworks.torch.pointers.pointer_tensor import PointerTensor
from syft.frameworks.torch.pointers.object_wrapper import ObjectWrapper

__all__ = ["ObjectPointer", "CallablePointer", "create_callable_pointer", "PointerTensor"]



import pytest

from unittest import mock

import torch
import torch.nn as nn
import torch.nn.functional as F
import syft as sy

import time
from syft.workers import WebsocketClientWorker
from syft.workers import WebsocketServerWorker
from syft.frameworks.torch.federated import utils

PRINT_IN_UNITTESTS = False


@pytest.mark.skip(reason="fails currently as it needs functions as torch.nn.linear to be unhooked.")
def test_train_config_with_jit_script_module(hook, workers):  # pragma: no cover
    alice = workers["alice"]
    me = workers["me"]

    data = torch.tensor([[-1, 2.0], [0, 1.1], [-1, 2.1], [0, 1.2]], requires_grad=True)
    target = torch.tensor([[1], [0], [1], [0]])

    dataset = sy.BaseDataset(data, target)
    alice.add_dataset(dataset, key="vectors")

    @hook.torch.jit.script
    def loss_fn(real, pred):
        return ((real.float() - pred.float()) ** 2).mean()

    # Model
    class Net(torch.jit.ScriptModule):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)
            self.fc3 = nn.Linear(2, 1)

        @torch.jit.script_method
        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = F.relu(self.fc2(x))
            x = self.fc3(x)
            return x

    model = Net()
    model.id = sy.ID_PROVIDER.pop()

    loss_fn.id = sy.ID_PROVIDER.pop()

    model_ptr = me.send(model, alice)
    loss_fn_ptr = me.send(loss_fn, alice)

    # Create and send train config
    train_config = sy.TrainConfig(model=model, loss_fn=loss_fn, batch_size=2)
    train_config.send(alice)

    for epoch in range(5):
        loss = alice.fit(dataset="vectors")
        if PRINT_IN_UNITTESTS:  # pragma: no cover:
            print("-" * 50)
            print("Iteration %s: alice's loss: %s" % (epoch, loss))

    print(alice)
    new_model = model_ptr.get()
    data = torch.tensor([[-1, 2.0], [0, 1.1], [-1, 2.1], [0, 1.2]], requires_grad=True)
    target = torch.tensor([[1], [0], [1], [0]])

    pred = model(data)
    loss_before = loss_fn(real=target, pred=pred)

    pred = new_model(data)
    loss_after = loss_fn(real=target, pred=pred)

    if PRINT_IN_UNITTESTS:  # pragma: no cover:
        print("Loss before training: {}".format(loss_before))
        print("Loss after training: {}".format(loss_after))

    assert loss_after < loss_before


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_train_config_with_jit_trace(hook, workers):  # pragma: no cover
    alice = workers["alice"]

    data = torch.tensor([[-1, 2.0], [0, 1.1], [-1, 2.1], [0, 1.2]], requires_grad=True)
    target = torch.tensor([[1], [0], [1], [0]])

    dataset = sy.BaseDataset(data, target)
    alice.add_dataset(dataset, key="gaussian_mixture")

    @hook.torch.jit.script
    def loss_fn(pred, target):
        return ((target.float() - pred.float()) ** 2).mean()

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)
            self.fc3 = nn.Linear(2, 1)

        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = F.relu(self.fc2(x))
            x = self.fc3(x)
            return x

    model_untraced = Net()

    model = torch.jit.trace(model_untraced, data)

    print("Evaluation before training")
    pred = model(data)
    loss_before = loss_fn(target=target, pred=pred)
    print("Loss: {}".format(loss_before))

    # Create and send train config
    train_config = sy.TrainConfig(model=model, loss_fn=loss_fn, batch_size=2)
    train_config.send(alice)

    for epoch in range(5):
        loss = alice.fit(dataset_key="gaussian_mixture")
        if PRINT_IN_UNITTESTS:  # pragma: no cover:
            print("-" * 50)
            print("Iteration %s: alice's loss: %s" % (epoch, loss))

    new_model = train_config.model_ptr.get()
    pred = new_model.obj(data)
    loss_after = loss_fn(target=target, pred=pred)

    if PRINT_IN_UNITTESTS:  # pragma: no cover:
        print("Loss before training: {}".format(loss_before))
        print("Loss after training: {}".format(loss_after))

    assert loss_after < loss_before


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_train_config_with_jit_trace_send_twice_with_fit(hook, workers):  # pragma: no cover
    alice = workers["alice"]
    model, loss_fn, data, target, loss_before, dataset_key = prepare_training(hook, alice)

    # Create and send train config
    train_config_0 = sy.TrainConfig(model=model, loss_fn=loss_fn, batch_size=2)
    train_config_0.send(alice)

    for epoch in range(5):
        loss = alice.fit(dataset_key=dataset_key)
        if PRINT_IN_UNITTESTS:  # pragma: no cover:
            print("-" * 50)
            print("TrainConfig 0, iteration %s: alice's loss: %s" % (epoch, loss))

    new_model = train_config_0.model_ptr.get()
    pred = new_model.obj(data)
    loss_after_0 = loss_fn(pred=pred, target=target)

    assert loss_after_0 < loss_before

    train_config = sy.TrainConfig(model=model, loss_fn=loss_fn, batch_size=2)
    train_config.send(alice)

    for epoch in range(5):
        loss = alice.fit(dataset_key=dataset_key)

        if PRINT_IN_UNITTESTS:  # pragma: no cover:
            print("-" * 50)
            print("TrainConfig 1, iteration %s: alice's loss: %s" % (epoch, loss))

    new_model = train_config.model_ptr.get()
    pred = new_model.obj(data)
    loss_after = loss_fn(pred=pred, target=target)
    if PRINT_IN_UNITTESTS:  # pragma: no cover:
        print("Loss after training with TrainConfig 0: {}".format(loss_after_0))
        print("Loss after training with TrainConfig 1:   {}".format(loss_after))

    assert loss_after < loss_before


def prepare_training(hook, alice):  # pragma: no cover

    data, target = utils.create_gaussian_mixture_toy_data(nr_samples=100)
    dataset_key = "gaussian_mixture"

    dataset = sy.BaseDataset(data, target)
    alice.add_dataset(dataset, key=dataset_key)

    @hook.torch.jit.script
    def loss_fn(pred, target):
        return ((target.float() - pred.float()) ** 2).mean()

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)
            self.fc3 = nn.Linear(2, 1)

            nn.init.xavier_uniform_(self.fc1.weight)
            nn.init.xavier_uniform_(self.fc2.weight)
            nn.init.xavier_uniform_(self.fc3.weight)

        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = F.relu(self.fc2(x))
            x = self.fc3(x)
            return x

    model_untraced = Net()

    model = torch.jit.trace(model_untraced, data)

    pred = model(data)
    loss_before = loss_fn(target=target, pred=pred)
    return model, loss_fn, data, target, loss_before, dataset_key


def test___str__():
    train_config = sy.TrainConfig(batch_size=2, id=99887766, model=None, loss_fn=None)

    train_config_str = str(train_config)
    str_expected = (
        "<TrainConfig id:99887766 owner:me epochs: 1 batch_size: 2 optimizer_args: {'lr': 0.1}>"
    )

    assert str_expected == train_config_str


def test_send(workers):
    alice = workers["alice"]

    train_config = sy.TrainConfig(batch_size=2, id="id", model=None, loss_fn=None)
    train_config.send(alice)

    assert alice.train_config.id == train_config.id
    assert alice.train_config._model_id == train_config._model_id
    assert alice.train_config._loss_fn_id == train_config._loss_fn_id
    assert alice.train_config.batch_size == train_config.batch_size
    assert alice.train_config.epochs == train_config.epochs
    assert alice.train_config.optimizer == train_config.optimizer
    assert alice.train_config.optimizer_args == train_config.optimizer_args
    assert alice.train_config.location == train_config.location


def test_send_model_and_loss_fn(workers):
    train_config = sy.TrainConfig(
        batch_size=2, id="send_model_and_loss_fn_tc", model=None, loss_fn=None
    )
    alice = workers["alice"]

    orig_func = sy.ID_PROVIDER.pop
    model_id = 44
    model_id_at_location = 44000
    loss_fn_id = 55
    loss_fn_id_at_location = 55000
    sy.ID_PROVIDER.pop = mock.Mock(
        side_effect=[model_id, model_id_at_location, loss_fn_id, loss_fn_id_at_location]
    )

    train_config.send(alice)

    assert alice.train_config.id == train_config.id
    assert alice.train_config._model_id == train_config._model_id
    assert alice.train_config._loss_fn_id == train_config._loss_fn_id
    assert alice.train_config.batch_size == train_config.batch_size
    assert alice.train_config.epochs == train_config.epochs
    assert alice.train_config.optimizer == train_config.optimizer
    assert alice.train_config.optimizer_args == train_config.optimizer_args
    assert alice.train_config.location == train_config.location
    assert alice.train_config._model_id == model_id
    assert alice.train_config._loss_fn_id == loss_fn_id

    sy.ID_PROVIDER.pop = orig_func


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
@pytest.mark.asyncio
async def test_train_config_with_jit_trace_async(hook, start_proc):  # pragma: no cover
    kwargs = {"id": "async_fit", "host": "localhost", "port": 8777, "hook": hook}
    # data = torch.tensor([[0.0, 1.0], [1.0, 0.0], [1.0, 1.0], [0.0, 0.0]], requires_grad=True)
    # target = torch.tensor([[1.0], [1.0], [0.0], [0.0]], requires_grad=False)
    # dataset_key = "xor"
    data, target = utils.create_gaussian_mixture_toy_data(100)
    dataset_key = "gaussian_mixture"

    mock_data = torch.zeros(1, 2)

    # TODO check reason for error (RuntimeError: This event loop is already running) when starting websocket server from pytest-asyncio environment
    # dataset = sy.BaseDataset(data, target)

    # process_remote_worker = start_proc(WebsocketServerWorker, dataset=(dataset, dataset_key), **kwargs)

    # time.sleep(0.1)

    local_worker = WebsocketClientWorker(**kwargs)

    @hook.torch.jit.script
    def loss_fn(pred, target):
        return ((target.view(pred.shape).float() - pred.float()) ** 2).mean()

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)
            self.fc3 = nn.Linear(2, 1)

        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = F.relu(self.fc2(x))
            x = self.fc3(x)
            return x

    model_untraced = Net()

    model = torch.jit.trace(model_untraced, mock_data)

    pred = model(data)
    loss_before = loss_fn(target=target, pred=pred)

    # Create and send train config
    train_config = sy.TrainConfig(
        model=model, loss_fn=loss_fn, batch_size=2, optimizer="SGD", optimizer_args={"lr": 0.1}
    )
    train_config.send(local_worker)

    for epoch in range(5):
        loss = await local_worker.async_fit(dataset_key=dataset_key)
        if PRINT_IN_UNITTESTS:  # pragma: no cover
            print("-" * 50)
            print("Iteration %s: alice's loss: %s" % (epoch, loss))

    new_model = train_config.model_ptr.get()

    assert not (model.fc1._parameters["weight"] == new_model.obj.fc1._parameters["weight"]).all()
    assert not (model.fc2._parameters["weight"] == new_model.obj.fc2._parameters["weight"]).all()
    assert not (model.fc3._parameters["weight"] == new_model.obj.fc3._parameters["weight"]).all()
    assert not (model.fc1._parameters["bias"] == new_model.obj.fc1._parameters["bias"]).all()
    assert not (model.fc2._parameters["bias"] == new_model.obj.fc2._parameters["bias"]).all()
    assert not (model.fc3._parameters["bias"] == new_model.obj.fc3._parameters["bias"]).all()

    new_model.obj.eval()
    pred = new_model.obj(data)
    loss_after = loss_fn(target=target, pred=pred)
    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Loss before training: {}".format(loss_before))
        print("Loss after training: {}".format(loss_after))

    local_worker.ws.shutdown()
    # process_remote_worker.terminate()

    assert loss_after < loss_before


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_train_config_with_jit_trace_sync(hook, start_proc):  # pragma: no cover
    kwargs = {"id": "sync_fit", "host": "localhost", "port": 9000, "hook": hook}

    data, target = utils.create_gaussian_mixture_toy_data(100)

    dataset = sy.BaseDataset(data, target)

    dataset_key = "gaussian_mixture"
    process_remote_worker = start_proc(
        WebsocketServerWorker, dataset=(dataset, dataset_key), **kwargs
    )

    time.sleep(0.1)

    local_worker = WebsocketClientWorker(**kwargs)

    @hook.torch.jit.script
    def loss_fn(pred, target):
        return ((target.view(pred.shape).float() - pred.float()) ** 2).mean()

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)
            self.fc3 = nn.Linear(2, 1)

        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = F.relu(self.fc2(x))
            x = self.fc3(x)
            return x

    model_untraced = Net()

    model = torch.jit.trace(model_untraced, data)

    pred = model(data)
    loss_before = loss_fn(pred=pred, target=target)

    # Create and send train config
    train_config = sy.TrainConfig(model=model, loss_fn=loss_fn, batch_size=2, epochs=1)
    train_config.send(local_worker)

    for epoch in range(5):
        loss = local_worker.fit(dataset_key=dataset_key)
        if PRINT_IN_UNITTESTS:  # pragma: no cover
            print("-" * 50)
            print("Iteration %s: alice's loss: %s" % (epoch, loss))

    new_model = train_config.model_ptr.get()

    # assert that the new model has updated (modified) parameters
    assert not (model.fc1._parameters["weight"] == new_model.obj.fc1._parameters["weight"]).all()
    assert not (model.fc2._parameters["weight"] == new_model.obj.fc2._parameters["weight"]).all()
    assert not (model.fc3._parameters["weight"] == new_model.obj.fc3._parameters["weight"]).all()
    assert not (model.fc1._parameters["bias"] == new_model.obj.fc1._parameters["bias"]).all()
    assert not (model.fc2._parameters["bias"] == new_model.obj.fc2._parameters["bias"]).all()
    assert not (model.fc3._parameters["bias"] == new_model.obj.fc3._parameters["bias"]).all()

    new_model.obj.eval()
    pred = new_model.obj(data)
    loss_after = loss_fn(pred=pred, target=target)

    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Loss before training: {}".format(loss_before))
        print("Loss after training: {}".format(loss_after))

    local_worker.ws.shutdown()
    del local_worker

    process_remote_worker.terminate()

    assert loss_after < loss_before



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""
This script computes bounds on the privacy cost of training the
student model from noisy aggregation of labels predicted by teachers.
It should be used only after training the student (and therefore the
teachers as well). We however include the label files required to
reproduce key results from our paper (https://arxiv.org/abs/1610.05755):
the epsilon bounds for MNIST and SVHN students.
"""
import os
import math
import numpy as np
import torch

# import tensorflow as tf
#
#
# # These parameters can be changed to compute bounds for different failure rates
# # or different model predictions.
#
# tf.flags.DEFINE_integer("moments",8, "Number of moments")
# tf.flags.DEFINE_float("noise_eps", 0.1, "Eps value for each call to noisymax.")
# tf.flags.DEFINE_float("delta", 1e-5, "Target value of delta.")
# tf.flags.DEFINE_float("beta", 0.09, "Value of beta for smooth sensitivity")
# tf.flags.DEFINE_string("counts_file","","Numpy matrix with raw counts")
# tf.flags.DEFINE_string("indices_file","",
#     "File containting a numpy matrix with indices used."
#     "Optional. Use the first max_examples indices if this is not provided.")
# tf.flags.DEFINE_integer("max_examples",1000,
#     "Number of examples to use. We will use the first"
#     " max_examples many examples from the counts_file"
#     " or indices_file to do the privacy cost estimate")
# tf.flags.DEFINE_float("too_small", 1e-10, "Small threshold to avoid log of 0")
# tf.flags.DEFINE_bool("input_is_counts", False, "False if labels, True if counts")
#
# FLAGS = tf.flags.FLAGS


def compute_q_noisy_max(counts, noise_eps):
    """Returns ~ Pr[outcome != winner].

  Args:
    counts: a list of scores
    noise_eps: privacy parameter for noisy_max
  Returns:
    q: the probability that outcome is different from true winner.
  """
    # For noisy max, we only get an upper bound.
    # Pr[ j beats i*] \leq (2+gap(j,i*))/ 4 exp(gap(j,i*)
    # proof at http://mathoverflow.net/questions/66763/
    # tight-bounds-on-probability-of-sum-of-laplace-random-variables

    winner = np.argmax(counts)
    counts_normalized = noise_eps * (counts - counts[winner])

    counts_rest = np.array([counts_normalized[i] for i in range(len(counts)) if i != winner])
    q = 0.0
    for c in counts_rest:
        gap = -c

        q += (gap + 2.0) / (4.0 * math.exp(gap))

    return min(q, 1.0 - (1.0 / len(counts)))


def compute_q_noisy_max_approx(counts, noise_eps):
    """Returns ~ Pr[outcome != winner].

  Args:
    counts: a list of scores
    noise_eps: privacy parameter for noisy_max
  Returns:
    q: the probability that outcome is different from true winner.
  """
    # For noisy max, we only get an upper bound.
    # Pr[ j beats i*] \leq (2+gap(j,i*))/ 4 exp(gap(j,i*)
    # proof at http://mathoverflow.net/questions/66763/
    # tight-bounds-on-probability-of-sum-of-laplace-random-variables
    # This code uses an approximation that is faster and easier
    # to get local sensitivity bound on.

    winner = np.argmax(counts)
    counts_normalized = noise_eps * (counts - counts[winner])
    counts_rest = np.array([counts_normalized[i] for i in range(len(counts)) if i != winner])
    gap = -max(counts_rest)
    q = (len(counts) - 1) * (gap + 2.0) / (4.0 * math.exp(gap))
    return min(q, 1.0 - (1.0 / len(counts)))


def logmgf_exact(q, priv_eps, l):
    """Computes the logmgf value given q and privacy eps.

  The bound used is the min of three terms. The first term is from
  https://arxiv.org/pdf/1605.02065.pdf.
  The second term is based on the fact that when event has probability (1-q) for
  q close to zero, q can only change by exp(eps), which corresponds to a
  much smaller multiplicative change in (1-q)
  The third term comes directly from the privacy guarantee.
  Args:
    q: pr of non-optimal outcome
    priv_eps: eps parameter for DP
    l: moment to compute.
  Returns:
    Upper bound on logmgf
  """
    if q < 0.5:
        t_one = (1 - q) * math.pow((1 - q) / (1 - math.exp(priv_eps) * q), l)
        t_two = q * math.exp(priv_eps * l)
        t = t_one + t_two
        try:
            log_t = math.log(t)
        except ValueError:
            print("Got ValueError in math.log for values :" + str((q, priv_eps, l, t)))
            log_t = priv_eps * l
    else:
        log_t = priv_eps * l

    return min(0.5 * priv_eps * priv_eps * l * (l + 1), log_t, priv_eps * l)


def logmgf_from_counts(counts, noise_eps, l):
    """
  ReportNoisyMax mechanism with noise_eps with 2*noise_eps-DP
  in our setting where one count can go up by one and another
  can go down by 1.
  """

    q = compute_q_noisy_max(counts, noise_eps)
    return logmgf_exact(q, 2.0 * noise_eps, l)


def sens_at_k(counts, noise_eps, l, k):
    """Return sensitivity at distane k.

  Args:
    counts: an array of scores
    noise_eps: noise parameter used
    l: moment whose sensitivity is being computed
    k: distance
  Returns:
    sensitivity: at distance k
  """
    counts_sorted = sorted(counts, reverse=True)
    if 0.5 * noise_eps * l > 1:
        print("l too large to compute sensitivity")
        return 0
    # Now we can assume that at k, gap remains positive
    # or we have reached the point where logmgf_exact is
    # determined by the first term and ind of q.
    if counts[0] < counts[1] + k:
        return 0
    counts_sorted[0] -= k
    counts_sorted[1] += k
    val = logmgf_from_counts(counts_sorted, noise_eps, l)
    counts_sorted[0] -= 1
    counts_sorted[1] += 1
    val_changed = logmgf_from_counts(counts_sorted, noise_eps, l)
    return val_changed - val


def smoothed_sens(counts, noise_eps, l, beta):
    """Compute beta-smooth sensitivity.

  Args:
    counts: array of scors
    noise_eps: noise parameter
    l: moment of interest
    beta: smoothness parameter
  Returns:
    smooth_sensitivity: a beta smooth upper bound
  """
    k = 0
    smoothed_sensitivity = sens_at_k(counts, noise_eps, l, k)
    while k < max(counts):
        k += 1
        sensitivity_at_k = sens_at_k(counts, noise_eps, l, k)
        smoothed_sensitivity = max(smoothed_sensitivity, math.exp(-beta * k) * sensitivity_at_k)
        if sensitivity_at_k == 0.0:
            break
    return smoothed_sensitivity


def perform_analysis(teacher_preds, indices, noise_eps, delta=1e-5, moments=8, beta=0.09):
    """"Performs PATE analysis on predictions from teachers and combined predictions for student.

    Args:
        teacher_preds: a numpy array of dim (num_teachers x num_examples). Each value corresponds to the
            index of the label which a teacher gave for a specific example
        indices: a numpy array of dim (num_examples) of aggregated examples which were aggregated using
            the noisy max mechanism.
        noise_eps: the epsilon level used to create the indices
        delta: the desired level of delta
        moments: the number of moments to track (see the paper)
        beta: a smoothing parameter (see the paper)
    Returns:
        tuple: first value is the data dependent epsilon, then the data independent epsilon
    """

    num_teachers, num_examples = teacher_preds.shape
    _num_examples = indices.shape[0]
    labels = set(list(teacher_preds.flatten()))
    num_labels = len(labels)

    assert num_examples == _num_examples

    counts_mat = np.zeros((num_examples, num_labels))

    for i in range(num_examples):
        for j in range(num_teachers):
            counts_mat[i, int(teacher_preds[j, i])] += 1

    l_list = 1.0 + np.array(range(moments))

    total_log_mgf_nm = np.array([0.0 for _ in l_list])
    total_ss_nm = np.array([0.0 for _ in l_list])

    for i in indices:

        total_log_mgf_nm += np.array(
            [logmgf_from_counts(counts_mat[i], noise_eps, l) for l in l_list]
        )

        total_ss_nm += np.array([smoothed_sens(counts_mat[i], noise_eps, l, beta) for l in l_list])

    # We want delta = exp(alpha - eps l).
    # Solving gives eps = (alpha - ln (delta))/l

    eps_list_nm = (total_log_mgf_nm - math.log(delta)) / l_list

    # print("Epsilons (Noisy Max): " + str(eps_list_nm))
    # print("Smoothed sensitivities (Noisy Max): " + str(total_ss_nm / l_list))

    # If beta < eps / 2 ln (1/delta), then adding noise Lap(1) * 2 SS/eps
    # is eps,delta DP
    # Also if beta < eps / 2(gamma +1), then adding noise 2(gamma+1) SS eta / eps
    # where eta has density proportional to 1 / (1+|z|^gamma) is eps-DP
    # Both from Corolloary 2.4 in
    # http://www.cse.psu.edu/~ads22/pubs/NRS07/NRS07-full-draft-v1.pdf
    # Print the first one's scale

    ss_eps = 2.0 * beta * math.log(1 / delta)
    ss_scale = 2.0 / ss_eps
    # print("To get an " + str(ss_eps) + "-DP estimate of epsilon, ")
    # print("..add noise ~ " + str(ss_scale))
    # print("... times " + str(total_ss_nm / l_list))
    # print("Epsilon = " + str(min(eps_list_nm)) + ".")
    if min(eps_list_nm) == eps_list_nm[-1]:
        print(
            "Warning: May not have used enough values of l. Increase 'moments' variable and run again."
        )

    # Data independent bound, as mechanism is
    # 2*noise_eps DP.
    data_ind_log_mgf = np.array([0.0 for _ in l_list])
    data_ind_log_mgf += num_examples * np.array(
        [logmgf_exact(1.0, 2.0 * noise_eps, l) for l in l_list]
    )

    data_ind_eps_list = (data_ind_log_mgf - math.log(delta)) / l_list
    # print("Data independent bound = " + str(min(data_ind_eps_list)) + ".")

    return min(eps_list_nm), min(data_ind_eps_list)


def tensors_to_literals(tensor_list):
    """Converts list of torch tensors to list of integers/floats. Fix for not having the functionality which converts list of tensors to tensors
    
       Args:
           
           tensor_list[List]: List of torch tensors
           
       Returns:
           
           literal_list[List]: List of floats/integers
           
    """

    literal_list = []

    for tensor in tensor_list:
        literal_list.append(tensor.item())

    return literal_list


def logmgf_exact_torch(q, priv_eps, l):
    """Computes the logmgf value given q and privacy eps.
       The bound used is the min of three terms. The first term is from
       https://arxiv.org/pdf/1605.02065.pdf.
       The second term is based on the fact that when event has probability (1-q) for
       q close to zero, q can only change by exp(eps), which corresponds to a
       much smaller multiplicative change in (1-q)
       The third term comes directly from the privacy guarantee.
       Args:
            q: pr of non-optimal outcome
            priv_eps: eps parameter for DP
            l: moment to compute.
       Returns:
            Upper bound on logmgf
      """
    if q < 0.5:
        t_one = (1 - q) * math.pow((1 - q) / (1 - math.exp(priv_eps) * q), l)
        t_two = q * math.exp(priv_eps * l)
        t = t_one + t_two
        try:

            log_t = math.log(t)

        except ValueError:

            print("Got ValueError in math.log for values :" + str((q, priv_eps, l, t)))
            log_t = priv_eps * l
    else:

        log_t = priv_eps * l

    return min(0.5 * priv_eps * priv_eps * l * (l + 1), log_t, priv_eps * l)


def compute_q_noisy_max_torch(counts, noise_eps):
    """Returns ~ Pr[outcome != winner].
       Args:
           
          counts: a list of scores
          noise_eps: privacy parameter for noisy_max
          
       Returns:
           
          q: the probability that outcome is different from true winner.
          
    """

    if type(counts) != torch.tensor:

        counts = torch.tensor(tensors_to_literals(counts), dtype=torch.float)

    _, winner = counts.max(0)
    counts_normalized = noise_eps * (
        torch.tensor(counts, dtype=torch.float) - torch.tensor(counts[winner], dtype=torch.float)
    )

    counts_normalized = tensors_to_literals(counts_normalized)
    counts_rest = torch.tensor(
        [counts_normalized[i] for i in range(len(counts)) if i != winner], dtype=torch.float
    )
    q = 0.0

    index = 0
    for c in counts_rest:

        gap = -c
        q += (gap + 2.0) / (4.0 * math.exp(gap))

        index += 1

    return min(q, 1.0 - (1.0 / len(counts)))


def logmgf_from_counts_torch(counts, noise_eps, l):

    """
        ReportNoisyMax mechanism with noise_eps with 2*noise_eps-DP
        in our setting where one count can go up by one and another
        can go down by 1.
    """

    q = compute_q_noisy_max_torch(counts, noise_eps)

    return logmgf_exact_torch(q, 2.0 * noise_eps, l)


def sens_at_k_torch(counts, noise_eps, l, k):

    """Return sensitivity at distane k.
      Args:
        
          counts: an array of scores
          noise_eps: noise parameter used
          l: moment whose sensitivity is being computed
          k: distance
      Returns:
         sensitivity: at distance k
     """

    counts_sorted = sorted(counts, reverse=True)

    if 0.5 * noise_eps * l > 1:

        print("l too large to compute sensitivity")
        return 0

    if counts[0] < counts[1] + k:

        return 0

    counts_sorted[0] -= k
    counts_sorted[1] += k
    val = logmgf_from_counts_torch(counts_sorted, noise_eps, l)
    counts_sorted[0] -= 1
    counts_sorted[1] += 1
    val_changed = logmgf_from_counts_torch(counts_sorted, noise_eps, l)
    return val_changed - val


def smooth_sens_torch(counts, noise_eps, l, beta):

    """Compute beta-smooth sensitivity.
    
     Args:
         counts: array of scors
         noise_eps: noise parameter
         l: moment of interest
         beta: smoothness parameter
     Returns:
         smooth_sensitivity: a beta smooth upper bound
     """

    k = 0
    smoothed_sensitivity = sens_at_k_torch(counts, noise_eps, l, k)

    while k < max(counts):

        k += 1
        sensitivity_at_k = sens_at_k_torch(counts, noise_eps, l, k)
        smoothed_sensitivity = max(smoothed_sensitivity, math.exp(-beta * k) * sensitivity_at_k)
        if sensitivity_at_k == 0.0:
            break

    return smoothed_sensitivity


def perform_analysis_torch(preds, indices, noise_eps=0.1, delta=1e-5, moments=8, beta=0.09):
    """Performs PATE analysis on predictions from teachers and combined predictions for student.
    Args:
        teacher_preds: a torch tensor of dim (num_teachers x num_examples). Each value corresponds to the
            index of the label which a teacher gave for a specific example
        indices: a torch tensor of dim (num_examples) of aggregated examples which were aggregated using
            the noisy max mechanism.
        noise_eps: the epsilon level used to create the indices
        delta: the desired level of delta
        moments: the number of moments to track (see the paper)
        beta: a smoothing parameter (see the paper)
    Returns:
        tuple: first value is the data dependent epsilon, then the data independent epsilon
    """

    num_teachers, num_examples = preds.shape
    _num_examples = indices.shape[0]

    assert num_examples == _num_examples

    labels = list(preds.flatten())
    labels = set([tensor.item() for tensor in labels])
    num_labels = len(labels)

    counts_mat = torch.zeros(num_examples, num_labels, dtype=torch.float32)

    for i in range(num_examples):

        for j in range(num_teachers):

            counts_mat[i, int(preds[j, i])] += 1

    l_list = 1 + torch.tensor(range(moments), dtype=torch.float)

    total_log_mgf_nm = torch.tensor([0.0 for _ in l_list], dtype=torch.float)
    total_ss_nm = torch.tensor([0.0 for _ in l_list], dtype=torch.float)

    for i in indices:

        total_log_mgf_nm += torch.tensor(
            [logmgf_from_counts_torch(counts_mat[i].clone(), noise_eps, l) for l in l_list]
        )

        total_ss_nm += torch.tensor(
            [smooth_sens_torch(counts_mat[i].clone(), noise_eps, l, beta) for l in l_list],
            dtype=torch.float,
        )

    eps_list_nm = (total_log_mgf_nm - math.log(delta)) / l_list
    ss_eps = 2.0 * beta * math.log(1 / delta)
    ss_scale = 2.0 / ss_eps
    if min(eps_list_nm) == eps_list_nm[-1]:
        print(
            "Warning: May not have used enough values of l. Increase 'moments' variable and run again."
        )

    data_ind_log_mgf = torch.tensor([0.0 for _ in l_list])
    data_ind_log_mgf += num_examples * torch.tensor(
        tensors_to_literals([logmgf_exact_torch(1.0, 2.0 * noise_eps, l) for l in l_list])
    )

    data_ind_eps_list = (data_ind_log_mgf - math.log(delta)) / l_list

    return min(eps_list_nm), min(data_ind_eps_list)



from syft.messaging.message import Message

from syft.messaging.plan import Plan
from syft.messaging.plan import func2plan
from syft.messaging.plan import method2plan
from syft.messaging.plan import make_plan



import io
from os.path import exists, join
import time
from socket import gethostname
from OpenSSL import crypto, SSL
import pytest
import torch
import syft as sy
from syft.frameworks.torch.federated import utils

from syft.workers import WebsocketClientWorker
from syft.workers import WebsocketServerWorker

PRINT_IN_UNITTESTS = False


def instantiate_websocket_client_worker(**kwargs):  # pragma: no cover
    """ Helper function to instantiate the websocket client.
    If connection is refused, we wait a bit and try again.
    After 5 failed tries, a ConnectionRefusedError is raised.
    """
    retry_counter = 0
    connection_open = False
    while not connection_open:
        try:
            local_worker = WebsocketClientWorker(**kwargs)
            connection_open = True
        except ConnectionRefusedError as e:
            if retry_counter < 5:
                retry_counter += 1
                time.sleep(0.1)
            else:
                raise e
    return local_worker


@pytest.mark.parametrize("secure", [True, False])
def test_websocket_worker_basic(hook, start_proc, secure, tmpdir):
    """Evaluates that you can do basic tensor operations using
    WebsocketServerWorker in insecure and secure mode."""

    def create_self_signed_cert(cert_path, key_path):
        # create a key pair
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 1024)

        # create a self-signed cert
        cert = crypto.X509()
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(1000)
        cert.set_pubkey(k)
        cert.sign(k, "sha1")

        # store keys and cert
        open(cert_path, "wb").write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
        open(key_path, "wb").write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k))

    kwargs = {
        "id": "secure_fed" if secure else "fed",
        "host": "localhost",
        "port": 8766 if secure else 8765,
        "hook": hook,
    }

    if secure:
        # Create cert and keys
        cert_path = tmpdir.join("test.crt")
        key_path = tmpdir.join("test.key")
        create_self_signed_cert(cert_path, key_path)
        kwargs["cert_path"] = cert_path
        kwargs["key_path"] = key_path

    process_remote_worker = start_proc(WebsocketServerWorker, **kwargs)

    time.sleep(0.1)
    x = torch.ones(5)

    if secure:
        # unused args
        del kwargs["cert_path"]
        del kwargs["key_path"]

    kwargs["secure"] = secure
    local_worker = instantiate_websocket_client_worker(**kwargs)

    x = x.send(local_worker)
    y = x + x
    y = y.get()

    assert (y == torch.ones(5) * 2).all()

    del x

    local_worker.ws.shutdown()
    time.sleep(0.1)
    local_worker.remove_worker_from_local_worker_registry()
    process_remote_worker.terminate()


def test_websocket_workers_search(hook, start_proc):
    """Evaluates that a client can search and find tensors that belong
    to another party"""
    # Sample tensor to store on the server
    sample_data = torch.tensor([1, 2, 3, 4]).tag("#sample_data", "#another_tag")
    # Args for initializing the websocket server and client
    base_kwargs = {"id": "fed2", "host": "localhost", "port": 8767, "hook": hook}
    server_kwargs = base_kwargs
    server_kwargs["data"] = [sample_data]
    process_remote_worker = start_proc(WebsocketServerWorker, **server_kwargs)

    local_worker = instantiate_websocket_client_worker(**base_kwargs)

    # Search for the tensor located on the server by using its tag
    results = local_worker.search("#sample_data", "#another_tag")

    assert results
    assert results[0].owner.id == "me"
    assert results[0].location.id == "fed2"

    # Search multiple times should still work
    results = local_worker.search("#sample_data", "#another_tag")

    assert results
    assert results[0].owner.id == "me"
    assert results[0].location.id == "fed2"

    local_worker.close()
    time.sleep(0.1)
    local_worker.remove_worker_from_local_worker_registry()
    process_remote_worker.terminate()


def test_list_objects_remote(hook, start_proc):

    kwargs = {"id": "fed", "host": "localhost", "port": 8765, "hook": hook}
    process_remote_fed1 = start_proc(WebsocketServerWorker, **kwargs)
    local_worker = instantiate_websocket_client_worker(**kwargs)

    x = torch.tensor([1, 2, 3]).send(local_worker)

    res = local_worker.list_objects_remote()
    res_dict = eval(res.replace("tensor", "torch.tensor"))
    assert len(res_dict) == 1

    y = torch.tensor([4, 5, 6]).send(local_worker)
    res = local_worker.list_objects_remote()
    res_dict = eval(res.replace("tensor", "torch.tensor"))
    assert len(res_dict) == 2

    # retrieve x and y before terminating the websocket connection
    x.get()
    y.get()
    local_worker.close()
    local_worker.remove_worker_from_local_worker_registry()
    process_remote_fed1.terminate()


def test_objects_count_remote(hook, start_proc):

    kwargs = {"id": "fed", "host": "localhost", "port": 8764, "hook": hook}
    process_remote_worker = start_proc(WebsocketServerWorker, **kwargs)
    local_worker = instantiate_websocket_client_worker(**kwargs)

    x = torch.tensor([1, 2, 3]).send(local_worker)

    nr_objects = local_worker.objects_count_remote()
    assert nr_objects == 1

    y = torch.tensor([4, 5, 6]).send(local_worker)
    nr_objects = local_worker.objects_count_remote()
    assert nr_objects == 2

    x.get()
    nr_objects = local_worker.objects_count_remote()
    assert nr_objects == 1

    # get remote object before terminating the websocket connection
    y.get()
    local_worker.close()
    local_worker.remove_worker_from_local_worker_registry()
    process_remote_worker.terminate()


def test_clear_objects_remote(hook, start_proc):
    kwargs = {"id": "fed", "host": "localhost", "port": 8769, "hook": hook}
    process_remote_worker = start_proc(WebsocketServerWorker, **kwargs)
    local_worker = instantiate_websocket_client_worker(**kwargs)

    x = torch.tensor([1, 2, 3]).send(local_worker, garbage_collect_data=False)
    y = torch.tensor(4).send(local_worker, garbage_collect_data=False)

    nr_objects = local_worker.objects_count_remote()
    assert nr_objects == 2

    local_worker.clear_objects_remote()
    nr_objects = local_worker.objects_count_remote()
    assert nr_objects == 0

    local_worker.close()
    local_worker.remove_worker_from_local_worker_registry()
    process_remote_worker.terminate()


def test_connect_close(hook, start_proc):
    kwargs = {"id": "fed", "host": "localhost", "port": 8765, "hook": hook}
    process_remote_worker = start_proc(WebsocketServerWorker, **kwargs)
    local_worker = instantiate_websocket_client_worker(**kwargs)

    x = torch.tensor([1, 2, 3])
    x_ptr = x.send(local_worker)

    assert local_worker.objects_count_remote() == 1

    local_worker.close()

    time.sleep(0.1)

    local_worker.connect()

    assert local_worker.objects_count_remote() == 1

    x_val = x_ptr.get()
    assert (x_val == x).all()

    local_worker.close()
    local_worker.remove_worker_from_local_worker_registry()
    local_worker.close()

    time.sleep(0.1)

    process_remote_worker.terminate()


def test_websocket_worker_multiple_output_response(hook, start_proc):
    """Evaluates that you can do basic tensor operations using
    WebsocketServerWorker"""

    kwargs = {"id": "socket_multiple_output", "host": "localhost", "port": 8768, "hook": hook}
    process_remote_worker = start_proc(WebsocketServerWorker, **kwargs)
    local_worker = instantiate_websocket_client_worker(**kwargs)

    x = torch.tensor([1.0, 3, 2])
    x = x.send(local_worker)
    p1, p2 = torch.sort(x)
    x1, x2 = p1.get(), p2.get()

    assert (x1 == torch.tensor([1.0, 2, 3])).all()
    assert (x2 == torch.tensor([0, 2, 1])).all()

    x.get()  # retrieve remote object before closing the websocket connection

    local_worker.close()
    local_worker.remove_worker_from_local_worker_registry()
    process_remote_worker.terminate()


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_evaluate(hook, start_proc):  # pragma: no cover

    sy.local_worker.clear_objects()
    sy.frameworks.torch.hook.hook_args.hook_method_args_functions = {}
    sy.frameworks.torch.hook.hook_args.hook_method_response_functions = {}
    sy.frameworks.torch.hook.hook_args.get_tensor_type_functions = {}
    sy.frameworks.torch.hook.hook_args.register_response_functions = {}

    data, target = utils.iris_data_partial()

    dataset = sy.BaseDataset(data=data, targets=target)

    kwargs = {"id": "evaluate_remote", "host": "localhost", "port": 8780, "hook": hook}
    dataset_key = "iris"
    # TODO: check why unit test sometimes fails when WebsocketServerWorker is started from the unit test. Fails when run after test_federated_client.py
    # process_remote_worker = start_proc(WebsocketServerWorker, dataset=(dataset, dataset_key), verbose=True, **kwargs)

    local_worker = instantiate_websocket_client_worker(**kwargs)

    def loss_fn(pred, target):
        return torch.nn.functional.cross_entropy(input=pred, target=target)

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = torch.nn.Linear(4, 3)

            torch.nn.init.xavier_normal_(self.fc1.weight)

        def forward(self, x):
            x = torch.nn.functional.relu(self.fc1(x))
            return x

    model_untraced = Net()
    model = torch.jit.trace(model_untraced, data)
    loss_traced = torch.jit.trace(loss_fn, (torch.tensor([[0.3, 0.5, 0.2]]), torch.tensor([1])))

    pred = model(data)
    loss_before = loss_fn(target=target, pred=pred)
    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Loss: {}".format(loss_before))

    # Create and send train config
    train_config = sy.TrainConfig(
        batch_size=4,
        model=model,
        loss_fn=loss_traced,
        model_id=None,
        loss_fn_id=None,
        optimizer_args=None,
        epochs=1,
    )
    train_config.send(local_worker)

    result = local_worker.evaluate(
        dataset_key=dataset_key, calculate_histograms=True, nr_bins=3, calculate_loss=True
    )

    test_loss_before, correct_before, len_dataset, hist_pred_before, hist_target = result

    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Evaluation result before training: {}".format(result))

    assert len_dataset == 30
    assert (hist_target == [10, 10, 10]).all()

    local_worker.close()
    local_worker.remove_worker_from_local_worker_registry()
    # process_remote_worker.terminate()



from syft.frameworks.torch.hook import hook_args
from syft.frameworks.torch import pointers
import torch
import numpy as np


def test_build_rule_syft_tensors_and_pointers():
    pointer = pointers.PointerTensor(
        id=1000, location="location", owner="owner", garbage_collect_data=False
    )
    result = hook_args.build_rule(([torch.tensor([1, 2]), pointer], 42))
    assert result == ([1, 1], 0)


def test_build_rule_numpy():
    arr = np.array([2.0, 3.0, 4.0])
    result = hook_args.build_rule([arr, arr + 2, [2, 4, "string"]])
    assert result == [1, 1, [0, 0, 0]]



from typing import List
from typing import Union

import torch

from syft.frameworks.torch.tensors.interpreters import AbstractTensor


class ObjectStorage:
    """A storage of objects identifiable by their id.

    A wrapper object to a collection of objects where all objects
    are stored using their IDs as keys.
    """

    def __init__(self):
        self._objects = {}

    def register_obj(self, obj: object, obj_id: Union[str, int] = None):
        """Registers the specified object with the current worker node.

        Selects an id for the object, assigns a list of owners, and establishes
        whether it's a pointer or not. This method is generally not used by the
        client and is instead used by internal processes (hooks and workers).

        Args:
            obj: A torch Tensor or Variable object to be registered.
            obj_id (int or string): random integer between 0 and 1e10 or
            string uniquely identifying the object.
        """
        if obj_id is not None and hasattr(obj, "id"):
            obj.id = obj_id
        self.set_obj(obj)

    def de_register_obj(self, obj: object, _recurse_torch_objs: bool = True):
        """Deregisters the specified object.

        Deregister and remove attributes which are indicative of registration.

        Args:
            obj: A torch Tensor or Variable object to be deregistered.
            _recurse_torch_objs: A boolean indicating whether the object is
                more complex and needs to be explored. Is not supported at the
                moment.
        """
        if hasattr(obj, "id"):
            self.rm_obj(obj.id)
        if hasattr(obj, "_owner"):
            del obj._owner

    def get_obj(self, obj_id: Union[str, int]) -> object:
        """Returns the object from registry.

        Look up an object from the registry using its ID.

        Args:
            obj_id: A string or integer id of an object to look up.

        Returns:
            Object with id equals to `obj_id`.
        """

        try:
            obj = self._objects[obj_id]
        except KeyError as e:
            if obj_id not in self._objects:
                msg = 'Object "' + str(obj_id) + '" not found on worker!!!'
                msg += (
                    "You just tried to interact with an object ID:"
                    + str(obj_id)
                    + " on "
                    + str(self)
                    + " which does not exist!!! "
                )
                msg += (
                    "Use .send() and .get() on all your tensors to make sure they're"
                    "on the same machines. "
                    "If you think this tensor does exist, check the ._objects dictionary"
                    "on the worker and see for yourself!!! "
                    "The most common reason this error happens is because someone calls"
                    ".get() on the object's pointer without realizing it (which deletes "
                    "the remote object and sends it to the pointer). Check your code to "
                    "make sure you haven't already called .get() on this pointer!!!"
                )
                raise KeyError(msg)
            else:
                raise e

        return obj

    def set_obj(self, obj: Union[torch.Tensor, AbstractTensor]) -> None:
        """Adds an object to the registry of objects.

        Args:
            obj: A torch or syft tensor with an id.
        """
        self._objects[obj.id] = obj

    def rm_obj(self, remote_key: Union[str, int]):
        """Removes an object.

        Remove the object from the permanent object registry if it exists.

        Args:
            remote_key: A string or integer representing id of the object to be
                removed.
        """
        if remote_key in self._objects:
            del self._objects[remote_key]

    def force_rm_obj(self, remote_key: Union[str, int]):
        """Forces object removal.

        Besides removing the object from the permanent object registry if it exists.
        Explicitly forces removal of the object modifying the `garbage_collect_data` attribute.

        Args:
            remote_key: A string or integer representing id of the object to be
                removed.
        """
        if remote_key in self._objects:
            obj = self._objects[remote_key]
            if hasattr(obj, "child"):
                obj.child.garbage_collect_data = True
            del self._objects[remote_key]

    def clear_objects(self, return_self: bool = True):
        """Removes all objects from the object storage.

        Note: the "return self" statement is kept in order to avoid modifying the code shown in the udacity course.

        Args:
            return_self: flag, whether to return self as return value

        Returns:
            self, if return_self if True, else None

        """
        self._objects.clear()
        return self if return_self else None

    def current_objects(self):
        """Returns a copy of the objects in the object storage."""
        return self._objects.copy()



from abc import ABC
import functools
import torch
from typing import List

import syft as sy
import weakref


class AbstractObject(ABC):
    """
    This is a generic object abstraction.
    """

    is_wrapper = False

    def __init__(
        self,
        id: int = None,
        owner: "sy.workers.AbstractWorker" = None,
        tags: List[str] = None,
        description: str = None,
        child=None,
    ):
        """Initializer for AbstractTensor

        Args:
            id: An optional string or integer id of the tensor
            owner: An optional BaseWorker object to specify the worker on which
                the tensor is located.
            tags: an optional set of hashtags corresponding to this tensor
                which this tensor should be searchable for
            description: an optional string describing the purpose of the
                tensor
            child: an optional tensor to put in the .child attribute to build
                a chain of tensors
        """
        self.owner = owner
        if id is None:
            self.id = sy.ID_PROVIDER.pop()
        else:
            self.id = id
        self.tags = tags
        self.description = description
        self.child = child

    def __str__(self) -> str:
        if hasattr(self, "child"):
            return type(self).__name__ + ">" + self.child.__str__()
        else:
            return type(self).__name__

    def __repr__(self) -> str:
        if hasattr(self, "child"):
            return type(self).__name__ + ">" + self.child.__repr__()
        else:
            return type(self).__name__

    def __len__(self) -> int:
        """Alias .shape[0] with len(), helpful for pointers"""
        try:
            if hasattr(self, "child") and not isinstance(self.child, dict):
                return self.child.shape[0]
            else:
                return self.shape[0]
        except IndexError:
            return 0

    @property
    def shape(self):
        return self.child.shape

    def serialize(self):  # check serde.py to see how to provide compression schemes
        """Serializes the tensor on which it's called.

        This is the high level convenience function for serializing torch
        tensors. It includes three steps, Simplify, Serialize, and Compress as
        described in serde.py.
        By default serde is compressing using LZ4

        Returns:
            The serialized form of the tensor.
            For example:
                x = torch.Tensor([1,2,3,4,5])
                x.serialize() # returns a serialized object
        """
        return sy.serde.serialize(self)

    def ser(self, *args, **kwargs):
        return self.serialize(*args, **kwargs)

    def get(self):
        """Just a pass through. This is most commonly used when calling .get() on a
        Syft tensor which has a child which is a pointer, an additive shared tensor,
        a multi-pointer, etc."""
        class_attributes = self.get_class_attributes()
        return type(self)(
            **class_attributes,
            owner=self.owner,
            tags=self.tags,
            description=self.description,
            id=self.id,
        ).on(self.child.get())

    def mid_get(self):
        """This method calls .get() on a child pointer and correctly registers the results"""

        child_id = self.id
        tensor = self.get()
        tensor.id = child_id
        self.owner.register_obj(tensor)

    def get_class_attributes(self):
        """
        Return all elements which defines an instance of a certain class.
        By default there is nothing so we return an empty dict, but for
        example for fixed precision tensor, the fractional precision is
        very important.
        """
        return {}

    @classmethod
    def on_function_call(cls, *args):
        """
        Override this to perform a specific action for each call of a torch
        function with arguments containing syft tensors of the class doing
        the overloading
        """
        pass

    @classmethod
    def handle_func_command(cls, command):
        """
        Receive an instruction for a function to be applied on a Syft Tensor,
        Replace in the args all the LogTensors with
        their child attribute, forward the command instruction to the
        handle_function_command of the type of the child attributes, get the
        response and replace a Syft Tensor on top of all tensors found in
        the response.

        Args:
            command: instruction of a function command: (command name,
            <no self>, arguments[, kwargs])

        Returns:
            the response of the function command
        """
        cmd, _, args, kwargs = command

        # Check that the function has not been overwritten
        try:
            # Try to get recursively the attributes in cmd = "<attr1>.<attr2>.<attr3>..."
            cmd = cls.rgetattr(cls, cmd)
            return cmd(*args, **kwargs)
        except AttributeError:
            pass

        # TODO: I can't manage the import issue, can you?
        # Replace all LoggingTensor with their child attribute
        new_args, new_kwargs, new_type = sy.frameworks.torch.hook_args.unwrap_args_from_function(
            cmd, args, kwargs
        )

        # build the new command
        new_command = (cmd, None, new_args, new_kwargs)

        # Do a generic action depending og the call
        cls.on_function_call(new_command)

        # Send it to the appropriate class and get the response
        response = new_type.handle_func_command(new_command)

        # Put back LoggingTensor on the tensors found in the response
        response = sy.frameworks.torch.hook_args.hook_response(cmd, response, wrap_type=cls)

        return response

    @classmethod
    def rgetattr(cls, obj, attr, *args):
        """
        Get an attribute recursively


        Args:
            obj: the object holding the attribute
            attr: nested attribute
            args: optional arguments to provide

        Returns:
            the attribute obj.attr

        Example:
            >>> rgetattr(obj, 'attr1.attr2.attr3')
            [Out] obj.attr1.attr2.attr3

        """

        def _getattr(obj, attr):
            return getattr(obj, attr, *args)

        return functools.reduce(_getattr, [obj] + attr.split("."))


class AbstractTensor(AbstractObject):
    def __init__(
        self,
        id: int = None,
        owner: "sy.workers.AbstractWorker" = None,
        tags: List[str] = None,
        description: str = None,
        child=None,
    ):
        super(AbstractTensor, self).__init__(id, owner, tags, description, child)

    def on(self, tensor: "AbstractTensor", wrap: bool = True) -> "AbstractTensor":
        """
        Add a syft(log) tensor on top of the tensor.

        Args:
            tensor: the tensor to extend
            wrap: if true, add the syft tensor between the wrapper
            and the rest of the chain. If false, just add it at the top

        Returns:
            a syft/torch tensor
        """
        if not wrap:
            self.child = tensor
            return self
        else:
            # if tensor is a wrapper
            if not hasattr(tensor, "child"):
                tensor = tensor.wrap()

            self.child = tensor.child
            tensor.child = self
            return tensor

    def wrap(self) -> torch.Tensor:
        """Wraps the class inside torch tensor.

        Because PyTorch does not (yet) support functionality for creating
        arbitrary Tensor types (via subclassing torch.Tensor), in order for our
        new tensor types (such as PointerTensor) to be usable by the rest of
        PyTorch (such as PyTorch's layers and loss functions), we need to wrap
        all of our new tensor types inside of a native PyTorch type.

        This function adds a .wrap() function to all of our tensor types (by
        adding it to AbstractTensor), such that (on any custom tensor
        my_tensor), my_tensor.wrap() will return a tensor that is compatible
        with the rest of the PyTorch API.

        Returns:
            A pytorch tensor.
        """
        wrapper = torch.Tensor()
        wrapper.child = self
        wrapper.is_wrapper = True
        wrapper.child.parent = weakref.ref(wrapper)

        if self.id is None:
            self.id = sy.ID_PROVIDER.pop()

        return wrapper

    def copy(self):
        return self + 0

    def refresh(self):
        """
        Forward to Additive Shared Tensor the call to refresh shares
        """
        if hasattr(self, "child"):
            self.child = self.child.refresh()
            return self
        else:
            raise AttributeError("Refresh should only be called on AdditiveSharedTensors")

    @property
    def grad(self):
        child_grad = self.child.grad
        if child_grad is None:
            return None
        else:
            return child_grad.wrap()


def initialize_tensor(
    hook_self, cls, torch_tensor: bool = False, owner=None, id=None, *init_args, **init_kwargs
):
    """Initializes the tensor.

    Args:
        hook_self: A reference to TorchHook class.
        cls: An object to keep track of id, owner and whether it is a native
            tensor or a wrapper over pytorch.
        torch_tensor: A boolean parameter (default False) to indicate whether
            it is torch tensor or not.
        owner: The owner of the tensor being initialised, leave it blank
            to if you have already provided a reference to TorchHook class.
        id: The id of tensor, a random id will be generated if there is no id
            specified.
    """
    cls.is_wrapper = False

    if not torch_tensor:
        cls.native___init__(*init_args, **init_kwargs)

    _apply_args(hook_self, cls, owner, id)


def _apply_args(hook_self, new_tensor, owner=None, id=None):

    if owner is None:
        owner = hook_self.local_worker

    if id is None:
        id = sy.ID_PROVIDER.pop()

    new_tensor.id = id
    new_tensor.owner = owner



import syft as sy
import torch
from typing import List
import logging

logger = logging.getLogger(__name__)


def extract_batches_per_worker(federated_train_loader: sy.FederatedDataLoader):
    """Extracts the batches from the federated_train_loader and stores them
       in a dictionary (keys = data.location).

       Args:
       federated_train_loader: the connection object we use to send responses.
                    back to the client.

    """
    logging_interval = 100
    batches = {}
    for worker_id in federated_train_loader.workers:
        worker = federated_train_loader.federated_dataset.datasets[worker_id].location
        batches[worker] = []

    for batch_idx, (data, target) in enumerate(federated_train_loader):
        if batch_idx % logging_interval == 0:
            logger.debug("Extracted %s batches from federated_train_loader", batch_idx)
        batches[data.location].append((data, target))

    return batches


def add_model(dst_model, src_model):
    """Add the parameters of two models.

        Args:
            dst_model (torch.nn.Module): the model to which the src_model will be added.
            src_model (torch.nn.Module): the model to be added to dst_model.
        Returns:
            torch.nn.Module: the resulting model of the addition.

        """

    params1 = src_model.named_parameters()
    params2 = dst_model.named_parameters()
    dict_params2 = dict(params2)
    with torch.no_grad():
        for name1, param1 in params1:
            if name1 in dict_params2:
                dict_params2[name1].set_(param1.data + dict_params2[name1].data)
    return dst_model


def scale_model(model, scale):
    """Scale the parameters of a model.

    Args:
        model (torch.nn.Module): the models whose parameters will be scaled.
        scale (float): the scaling factor.
    Returns:
        torch.nn.Module: the module with scaled parameters.

    """
    params = model.named_parameters()
    dict_params = dict(params)
    with torch.no_grad():
        for name, param in dict_params.items():
            dict_params[name].set_(dict_params[name].data * scale)
    return model


def federated_avg(models: List[torch.nn.Module]) -> torch.nn.Module:
    """Calculate the federated average of a list of models.

    Args:
        models (List[torch.nn.Module]): the models of which the federated average is calculated.

    Returns:
        torch.nn.Module: the module with averaged parameters.
    """
    nr_models = len(models)
    model_list = list(models.values())
    model = model_list[0]
    for i in range(1, nr_models):
        model = add_model(model, model_list[i])
    model = scale_model(model, 1.0 / nr_models)
    return model


def accuracy(pred_softmax, target):
    """Calculate the accuray of a given prediction.

    This functions assumes pred_softmax to be converted into the final prediction by taking the argmax.

    Args:
        pred_softmax: array type(float), providing nr_classes values per element in target.
        target: array type(int), correct classes, taking values in range [0, nr_classes).

    Returns:
        accuracy: float, fraction of correct predictions.

    """
    nr_elems = len(target)
    pred = pred_softmax.argmax(dim=1)
    return (pred.float() == target.view(pred.shape).float()).sum().numpy() / float(nr_elems)


def create_gaussian_mixture_toy_data(nr_samples: int):  # pragma: no cover
    """ Create a simple toy data for binary classification

    The data is drawn from two normal distributions
    target = 1: mu = 2, sigma = 1
    target = 0: mu = 0, sigma = 1
    The dataset is balanced with an equal number of positive and negative samples

    Args:
        nr_samples: number of samples to generate

    Returns:
        data, targets


    """
    sample_dim = 2
    one_half = int(nr_samples / 2)
    X1 = torch.randn(one_half, sample_dim, requires_grad=True)
    X2 = torch.randn(one_half, sample_dim, requires_grad=True) + 2
    X = torch.cat([X1, X2], dim=0)
    Y1 = torch.zeros(one_half, requires_grad=False).long()
    Y2 = torch.ones(one_half, requires_grad=False).long()
    Y = torch.cat([Y1, Y2], dim=0)
    return X, Y


def iris_data_partial():
    """

    Returns: 30 samples from the iris data set: https://archive.ics.uci.edu/ml/datasets/iris

    """
    data = [
        [5.1, 3.5, 1.4, 0.2],
        [4.9, 3.0, 1.4, 0.2],
        [4.7, 3.2, 1.3, 0.2],
        [4.6, 3.1, 1.5, 0.2],
        [5.0, 3.6, 1.4, 0.2],
        [5.4, 3.9, 1.7, 0.4],
        [4.6, 3.4, 1.4, 0.3],
        [5.0, 3.4, 1.5, 0.2],
        [4.4, 2.9, 1.4, 0.2],
        [4.9, 3.1, 1.5, 0.1],
    ]

    target_to_string = {0: "Iris-setosa", 1: "Iris-versicolor", 2: "Iris-virginica"}
    targets = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    data += [
        [7.0, 3.2, 4.7, 1.4],
        [6.4, 3.2, 4.5, 1.5],
        [6.9, 3.1, 4.9, 1.5],
        [5.5, 2.3, 4.0, 1.3],
        [6.5, 2.8, 4.6, 1.5],
        [5.7, 2.8, 4.5, 1.3],
        [6.3, 3.3, 4.7, 1.6],
        [4.9, 2.4, 3.3, 1.0],
        [6.6, 2.9, 4.6, 1.3],
        [5.2, 2.7, 3.9, 1.4],
    ]

    targets += [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]

    data += [
        [6.3, 3.3, 6.0, 2.5],
        [5.8, 2.7, 5.1, 1.9],
        [7.1, 3.0, 5.9, 2.1],
        [6.3, 2.9, 5.6, 1.8],
        [6.5, 3.0, 5.8, 2.2],
        [7.6, 3.0, 6.6, 2.1],
        [4.9, 2.5, 4.5, 1.7],
        [7.3, 2.9, 6.3, 1.8],
        [6.7, 2.5, 5.8, 1.8],
        [7.2, 3.6, 6.1, 2.5],
    ]

    targets += [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]

    return torch.tensor(data), torch.tensor(targets)



import syft as sy
import torch as th


def test_is_client_true(hook):
    me = hook.local_worker
    me.is_client_worker = True
    x = th.tensor([1, 2, 3])
    assert x.id not in me._objects


def test_is_client_false(hook):
    me = hook.local_worker
    me.is_client_worker = False
    x = th.tensor([1, 2, 3])
    assert x.id in me._objects


def test_in_known_workers(hook):
    # Get local worker
    local_worker = hook.local_worker
    id = local_worker.id

    # Get known workers dict
    known_workers = local_worker._known_workers

    assert id in known_workers and local_worker == known_workers[id]



from typing import Union
from typing import Callable
from typing import Any
from types import ModuleType


class TorchAttributes(object):
    """Adds torch module related custom attributes.

    TorchAttributes is a special class where all custom attributes related
    to the torch module can be added. Any global parameter, configuration,
    or reference relating to PyTorch should be stored here instead of
    attaching it directly to some other part of the global namespace.

    The main reason we need this is because the hooking process occasionally
    needs to save global objects, notably including what methods to hook and
    what methods to NOT hook.

    This will hold all necessary attributes PySyft needs.

    Args:
        torch: A ModuleType indicating the torch module
        hook: A ModuleType indicating the modules to hook
    """

    def __init__(self, torch: ModuleType, hook: ModuleType) -> None:
        """Initialization of the TorchAttributes class."""

        # SECTION: List all functions in torch module that we want to overload

        self.hook = hook

        # List modules that we will hook
        self.torch_modules = {
            "torch": torch,
            "torch.functional": torch.functional,
            "torch.nn.functional": torch.nn.functional,
        }

        # List all the function names with module as prefix in the modules to hook
        self.torch_modules_functions = {
            f"{module_name}.{func_name}"
            for module_name, torch_module in self.torch_modules.items()
            for func_name in dir(torch_module)
        }

        # Store reference to all torch functions by string name stored in torch_modules_functions
        self.eval_torch_modules_functions = {
            f"{module_name}.{func_name}": getattr(torch_module, func_name)
            for module_name, torch_module in self.torch_modules.items()
            for func_name in dir(torch_module)
        }

        # Add special functions to exclude from the hook **in alphabetical order**
        # Reasons can be:
        # - Used for internal process like printing tensors
        # - Don't use tensors so are bound to have local executions
        # - etc
        # DON'T put here:
        # - functions like native_*
        # - functions that could use pointers or syft tensors
        self.exclude = [
            "arange",
            "as_tensor",
            "from_numpy",
            "get_default_dtype",
            "get_device",
            "get_file_path",
            "get_num_threads",
            "get_rng_state",
            "int",
            "int16",
            "int32",
            "int64",
            "int8",
            "is_anomaly_enabled",
            "is_complex",
            "is_distributed",
            "is_floating_point",
            "is_grad_enabled",
            "is_nonzero",
            "is_same_size",
            "is_signed",
            "is_storage",
            "is_tensor",
            "isclose",
            "isfinite",
            "load",
            "manual_seed",
            "ones",
            "rand",
            "randint",
            "randn",
            "randn_like",
            "randperm",
            "range",
            "save",
            "set_",
            "set_num_threads",
            "short",
            "size",
            "storage",
            "storage_offset",
            "stride",
            "tensor",
            "typename",
            "zeros",
        ]

        # SECTION: List all torch tensor methods we want to overload
        self.tensor_types = [torch.Tensor]

        self.tensorvar_methods = list(
            {method for tensorvar in self.tensor_types for method in dir(tensorvar)}
        )
        self.tensorvar_methods += ["get_shape", "share", "fix_precision", "decode", "end_get"]

        # SECTION: Build the guard, that define which functions or methods can be safely called by
        # external or local workers

        # Add all tensor types
        self.guard = {
            "FloatTensor": torch.FloatTensor,
            "DoubleTensor": torch.DoubleTensor,
            "HalfTensor": torch.HalfTensor,
            "ByteTensor": torch.ByteTensor,
            "CharTensor": torch.CharTensor,
            "ShortTensor": torch.ShortTensor,
            "IntTensor": torch.IntTensor,
            "LongTensor": torch.LongTensor,
            "Parameter": torch.nn.Parameter,
        }

        # Allow the `syft.` prefix to be used
        keys = list(self.guard.keys())
        for key in keys:
            self.guard[f"syft.{key}"] = self.guard[key]

        # Concatenate torch functions and torch methods
        self.allowed_commands = {
            "tensorvar_methods": self.tensorvar_methods,
            "torch_modules": self.torch_modules_functions,
        }

        # The equivalent concatenation of native torch function names and native torch method names
        self.native_commands = {
            command_type: {cmd: self.get_native_torch_name(cmd) for cmd in commands}
            for command_type, commands in self.allowed_commands.items()
        }

        self.command_guard = self._command_guard

        # Dict {method_name: <is_inplace:bool>
        self.inplace_methods = {}

    def _command_guard(
        self, command: str, torch_domain: str, get_native: bool = False
    ) -> Union[Callable[..., Any], str]:
        """Checks command is in a given torch_domain and can be safely used.

        Args:
            command: A string indicating command name.
            torch_domain: A string indicating torch domain name or module in
                which the command is supposed to be.
            get_native: A boolean parameter (default False) to indicate whether
                to return the command name or the native torch function. If
                False, return command name else return the native torch
                function.

        Returns:
            The command name or a native torch function
        """
        if command not in self.allowed_commands[torch_domain]:
            raise RuntimeError(f'Command "{command}" is not a supported Torch operation.')
        if get_native:
            return self.native_commands[torch_domain][command]
        return command

    def _is_command_valid_guard(self, command: str, torch_domain: str) -> bool:
        """Validates the command.

        Indicates whether a command is valid with respect to the torch guard

        Args:
            command: A string indicating command to test.
            torch_domain: A string indicating the torch domain or module in
                which the command is supposed to be.

        Returns:
            A boolean indicating whether the command is valid.
        """
        try:
            self._command_guard(command, torch_domain)
        except RuntimeError:
            return False
        return True

    def eval_torch_modules(self) -> None:
        """Builds a mapping between the hooked and native commands.

        For each torch command functions in native_commands, transform the
        dictionary so that to each key, which is the name of the hooked
        command, now corresponds a value which is the evaluated native name of
        the command, namely the native command.

        Note that we don't do this for methods.
        """
        for cmd_name, native_cmd_name in self.native_commands["torch_modules"].items():
            if cmd_name not in self.torch_exclude:
                self.native_commands["torch_modules"][cmd_name] = self.eval_torch_modules_functions[
                    cmd_name
                ]

    @staticmethod
    def get_native_torch_name(attr: str) -> str:
        """Returns the name of the native command for the given hooked command.

        Args:
            attr: A string indicating the hooked command name (ex: torch.add)

        Returns:
            The name of the native command (ex: torch.native_add)
        """
        parts = attr.split(".")
        parts[-1] = "native_" + parts[-1]
        native_func_name = ".".join(parts)
        return native_func_name

    def is_inplace_method(self, method_name):
        """
        Says if a method is inplace or not by test if it ends by _ and is not a __xx__
        :param method_name: the name for the method
        :return: boolean
        """
        try:
            return self.inplace_methods[method_name]
        except KeyError:
            is_inplace = method_name[-1] == "_" and "__" not in method_name
            self.inplace_methods[method_name] = is_inplace
            return is_inplace

    @staticmethod
    def apply_fix16922(torch):
        """
        Apply the fix made in PR16922 of PyTorch until people use PyTorch 1.0.2
        :param torch: the pytorch module
        """
        broken_funcs = [
            "max_pool1d",
            "max_pool2d",
            "max_pool3d",
            "adaptive_max_pool1d",
            "adaptive_max_pool2d",
            "adaptive_max_pool3d",
        ]
        for broken_func in broken_funcs:
            getattr(torch.nn.functional, broken_func).__module__ = "torch.nn.functional"
            getattr(torch.nn.functional, broken_func).__name__ = broken_func



import logging
import syft as sy
from syft.workers import WebsocketServerWorker
import torch
import argparse
from torchvision import datasets
from torchvision import transforms
import numpy as np
from syft.frameworks.torch.federated import utils

KEEP_LABELS_DICT = {
    "alice": [0, 1, 2, 3],
    "bob": [4, 5, 6],
    "charlie": [7, 8, 9],
    "testing": list(range(10)),
}  # pragma: no cover


def start_websocket_server_worker(
    id, host, port, hook, verbose, keep_labels=None, training=True
):  # pragma: no cover
    """Helper function for spinning up a websocket server and setting up the local datasets."""

    server = WebsocketServerWorker(id=id, host=host, port=port, hook=hook, verbose=verbose)

    # Setup toy data (mnist example)
    mnist_dataset = datasets.MNIST(
        root="./data",
        train=training,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        ),
    )

    if training:
        indices = np.isin(mnist_dataset.targets, keep_labels).astype("uint8")
        logger.info("number of true indices: %s", indices.sum())
        selected_data = (
            torch.native_masked_select(mnist_dataset.data.transpose(0, 2), torch.tensor(indices))
            .view(28, 28, -1)
            .transpose(2, 0)
        )
        logger.info("after selection: %s", selected_data.shape)
        selected_targets = torch.native_masked_select(mnist_dataset.targets, torch.tensor(indices))

        dataset = sy.BaseDataset(
            data=selected_data, targets=selected_targets, transform=mnist_dataset.transform
        )
        key = "mnist"
    else:
        dataset = sy.BaseDataset(
            data=mnist_dataset.data,
            targets=mnist_dataset.targets,
            transform=mnist_dataset.transform,
        )
        key = "mnist_testing"

    server.add_dataset(dataset, key=key)

    # Setup toy data (vectors example)
    data_vectors = torch.tensor([[-1, 2.0], [0, 1.1], [-1, 2.1], [0, 1.2]], requires_grad=True)
    target_vectors = torch.tensor([[1], [0], [1], [0]])

    server.add_dataset(sy.BaseDataset(data_vectors, target_vectors), key="vectors")

    # Setup toy data (xor example)
    data_xor = torch.tensor([[0.0, 1.0], [1.0, 0.0], [1.0, 1.0], [0.0, 0.0]], requires_grad=True)
    target_xor = torch.tensor([1.0, 1.0, 0.0, 0.0], requires_grad=False)

    server.add_dataset(sy.BaseDataset(data_xor, target_xor), key="xor")

    # Setup gaussian mixture dataset
    data, target = utils.create_gaussian_mixture_toy_data(nr_samples=100)
    server.add_dataset(sy.BaseDataset(data, target), key="gaussian_mixture")

    # Setup partial iris dataset
    data, target = utils.iris_data_partial()
    dataset = sy.BaseDataset(data, target)
    dataset_key = "iris"
    server.add_dataset(dataset, key=dataset_key)

    logger.info("datasets: %s", server.datasets)
    if training:
        logger.info("len(datasets[mnist]): %s", len(server.datasets["mnist"]))

    server.start()
    return server


if __name__ == "__main__":  # pragma: no cover
    # Logging setup
    logger = logging.getLogger("run_websocket_server")
    FORMAT = "%(asctime)s %(levelname)s %(filename)s(l:%(lineno)d, p:%(process)d) - %(message)s"
    logging.basicConfig(format=FORMAT)
    logger.setLevel(level=logging.DEBUG)

    # Parse args
    parser = argparse.ArgumentParser(description="Run websocket server worker.")
    parser.add_argument(
        "--port",
        "-p",
        type=int,
        help="port number of the websocket server worker, e.g. --port 8777",
    )
    parser.add_argument("--host", type=str, default="localhost", help="host for the connection")
    parser.add_argument(
        "--id", type=str, help="name (id) of the websocket server worker, e.g. --id alice"
    )
    parser.add_argument(
        "--testing",
        action="store_true",
        help="if set, websocket server worker will load the test dataset instead of the training dataset",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="if set, websocket server worker will be started in verbose mode",
    )

    args = parser.parse_args()

    # Hook and start server
    hook = sy.TorchHook(torch)
    server = start_websocket_server_worker(
        id=args.id,
        host=args.host,
        port=args.port,
        hook=hook,
        verbose=args.verbose,
        keep_labels=KEEP_LABELS_DICT[args.id] if args.id in KEEP_LABELS_DICT else list(range(10)),
        training=not args.testing,
    )



import torch
from torch.utils.data import SequentialSampler, RandomSampler, BatchSampler
from torch._six import string_classes, int_classes, container_abcs

import logging
import math

numpy_type_map = {
    "float64": torch.DoubleTensor,
    "float32": torch.FloatTensor,
    "float16": torch.HalfTensor,
    "int64": torch.LongTensor,
    "int32": torch.IntTensor,
    "int16": torch.ShortTensor,
    "int8": torch.CharTensor,
    "uint8": torch.ByteTensor,
}


def default_collate(batch):
    """Puts each data field into a tensor with outer dimension batch size"""

    error_msg = "batch must contain tensors, numbers, dicts or lists; found {}"
    elem_type = type(batch[0])
    if isinstance(batch[0], torch.Tensor):
        return torch.stack(batch, 0)
    elif (
        elem_type.__module__ == "numpy"
        and elem_type.__name__ != "str_"
        and elem_type.__name__ != "string_"
    ):  # pragma: no cover
        elem = batch[0]
        if elem_type.__name__ == "ndarray":
            return torch.stack([torch.from_numpy(b) for b in batch], 0)
        if elem.shape == ():  # scalars
            py_type = float if elem.dtype.name.startswith("float") else int
            return numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
    elif isinstance(batch[0], int_classes):  # pragma: no cover
        return torch.LongTensor(batch)
    elif isinstance(batch[0], float):  # pragma: no cover
        return torch.DoubleTensor(batch)
    elif isinstance(batch[0], string_classes):  # pragma: no cover
        return batch
    elif isinstance(batch[0], container_abcs.Mapping):  # pragma: no cover
        return {key: default_collate([d[key] for d in batch]) for key in batch[0]}
    elif isinstance(batch[0], container_abcs.Sequence):  # pragma: no cover
        transposed = zip(*batch)
        return [default_collate(samples) for samples in transposed]

    raise TypeError((error_msg.format(type(batch[0]))))


class _DataLoaderIter(object):
    """Iterates once over the DataLoader's dataset, as specified by the samplers"""

    def __init__(self, loader, worker_idx):
        self.loader = loader
        self.federated_dataset = loader.federated_dataset

        # Assign the first worker to invoke
        self.worker_idx = worker_idx
        # List workers in a dict
        self.workers = {idx: worker for idx, worker in enumerate(loader.workers)}

        # The function used to stack all samples together
        self.collate_fn = loader.collate_fn

        # Create a sample iterator for each worker
        self.sample_iter = {
            worker: iter(batch_sampler) for worker, batch_sampler in loader.batch_samplers.items()
        }

    def __len__(self):
        return len(self.federated_dataset)

    def _get_batch(self):
        # If all workers have been used, end the iterator
        if len(self.workers) == 0:
            self.stop()

        worker = self.workers[self.worker_idx]

        try:
            indices = next(self.sample_iter[worker])
            batch = self.collate_fn([self.federated_dataset[worker][i] for i in indices])
            return batch
        # All the data for this worker has been used
        except StopIteration:
            # Forget this worker
            del self.workers[self.worker_idx]
            # Find another worker which is not busy
            worker_busy_ids = [it.worker_idx for it in self.loader.iterators]
            for idx in self.workers.keys():
                if idx not in worker_busy_ids:
                    self.worker_idx = idx
                    return self._get_batch()

            # If nothing is found, stop the iterator
            self.stop()

    def __next__(self):
        batch = self._get_batch()
        return batch

    def __iter__(self):
        return self

    def stop(self):
        self.worker_idx = -1
        raise StopIteration


class _DataLoaderOneWorkerIter(object):
    """Iterates once over the worker's dataset, as specified by its sampler"""

    def __init__(self, loader, worker_idx):
        self.loader = loader
        self.federated_dataset = loader.federated_dataset

        # Assign the worker to invoke
        self.worker = loader.workers[worker_idx]

        # The function used to stack all samples together
        self.collate_fn = loader.collate_fn

        # Create a sample iterator for each worker
        self.sample_iter = iter(loader.batch_samplers[self.worker])

    def _get_batch(self):
        # If all workers have been used, end the iterator
        if not self.worker:
            self.stop()

        try:
            indices = next(self.sample_iter)
            batch = self.collate_fn([self.federated_dataset[self.worker][i] for i in indices])
            return batch
        # All the data for this worker has been used
        except StopIteration:
            # If nothing is found, stop the iterator
            self.stop()

    # TODO: implement a length function. It should return the number of elements of the federated dataset that are
    #       located at this worker
    # def __len__(self):
    #    return len(self.federated_dataset)

    def __next__(self):
        return self._get_batch()

    def __iter__(self):
        return self

    def stop(self):
        self.worker = None
        raise StopIteration


class FederatedDataLoader(object):
    """
    Data loader. Combines a dataset and a sampler, and provides
    single or several iterators over the dataset.

    Arguments:
        federated_dataset (FederatedDataset): dataset from which to load the data.
        batch_size (int, optional): how many samples per batch to load
            (default: ``1``).
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: ``False``).
        collate_fn (callable, optional): merges a list of samples to form a mini-batch.
        drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
            if the dataset size is not divisible by the batch size. If ``False`` and
            the size of dataset is not divisible by the batch size, then the last batch
            will be smaller. (default: ``False``)
        num_iterators (int): number of workers from which to retrieve data in parallel.
            num_iterators <= len(federated_dataset.workers) - 1
            the effect is to retrieve num_iterators epochs of data but at each step data from num_iterators distinct
            workers is returned.
        iter_per_worker (bool): if set to true, __next__() will return a dictionary containing one batch per worker
    """

    __initialized = False

    def __init__(
        self,
        federated_dataset,
        batch_size=8,
        shuffle=False,
        num_iterators=1,
        drop_last=False,
        collate_fn=default_collate,
        iter_per_worker=False,
        **kwargs,
    ):
        if len(kwargs) > 0:
            options = ", ".join([f"{k}: {v}" for k, v in kwargs.items()])
            logging.warning(f"The following options are not supported: {options}")

        try:
            self.workers = federated_dataset.workers
        except AttributeError:
            raise Exception(
                "Your dataset is not a FederatedDataset, please use "
                "torch.utils.data.DataLoader instead."
            )

        self.federated_dataset = federated_dataset
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.collate_fn = collate_fn
        self.iter_class = _DataLoaderOneWorkerIter if iter_per_worker else _DataLoaderIter

        # Build a batch sampler per worker
        self.batch_samplers = {}
        for worker in self.workers:
            data_range = range(len(federated_dataset[worker]))
            if shuffle:
                sampler = RandomSampler(data_range)
            else:
                sampler = SequentialSampler(data_range)
            batch_sampler = BatchSampler(sampler, batch_size, drop_last)
            self.batch_samplers[worker] = batch_sampler

        if iter_per_worker:
            self.num_iterators = len(self.workers)
        else:
            # You can't have more iterators than n - 1 workers, because you always
            # need a worker idle in the worker switch process made by iterators
            if len(self.workers) == 1:
                self.num_iterators = 1
            else:
                self.num_iterators = min(num_iterators, len(self.workers) - 1)

    def __iter__(self):
        self.iterators = list()
        for idx in range(self.num_iterators):
            self.iterators.append(self.iter_class(self, worker_idx=idx))
        return self

    def __next__(self):
        if self.num_iterators > 1:
            batches = {}
            for iterator in self.iterators:
                data, target = next(iterator)
                batches[data.location] = (data, target)
            return batches
        else:
            iterator = self.iterators[0]
            data, target = next(iterator)
            return data, target

    def __len__(self):
        length = len(self.federated_dataset) / self.batch_size
        if self.drop_last:
            return int(length)
        else:
            return math.ceil(length)



"""All the tests relative to garbage collection of all kinds of remote or local tensors"""
import time

import torch

from syft.frameworks.torch.tensors.decorators import LoggingTensor
from syft.workers import WebsocketServerWorker
from syft.workers import WebsocketClientWorker

# TESTING POINTERS


def test_explicit_garbage_collect_pointer(workers):
    """Tests whether deleting a PointerTensor garbage collects the remote object too"""
    bob = workers["bob"]

    alice, bob = workers["alice"], workers["bob"]

    # create tensor
    x = torch.Tensor([1, 2])

    # send tensor to bob
    x_ptr = x.send(bob)

    # ensure bob has tensor
    assert x.id in bob._objects

    # delete pointer to tensor, which should
    # automatically garbage collect the remote
    # object on Bob's machine
    del x_ptr

    # ensure bob's object was garbage collected
    assert x.id not in bob._objects


def test_explicit_garbage_collect_double_pointer(workers):
    """Tests whether deleting a pointer to a pointer garbage collects
    the remote object too"""

    alice, bob = workers["alice"], workers["bob"]

    # create tensor
    x = torch.Tensor([1, 2])

    # send tensor to bob and then pointer to alice
    x_ptr = x.send(bob)
    x_ptr_ptr = x_ptr.send(alice)

    # ensure bob has tensor
    assert x.id in bob._objects

    # delete pointer to pointer to tensor, which should automatically
    # garbage collect the remote object on Bob's machine
    del x_ptr_ptr

    # ensure bob's object was garbage collected
    assert x.id not in bob._objects
    # TODO: shouldn't we check that alice's object was
    # garbage collected as well?
    # assert x.id not in workers["alice"]._objects

    # Chained version
    x = torch.Tensor([1, 2])
    x_id = x.id
    # send tensor to bob and then pointer to alice
    x = x.send(bob).send(alice)
    # ensure bob has tensor
    assert x_id in bob._objects
    # delete pointer to pointer to tensor
    del x
    # ensure bob's object was garbage collected
    assert x_id not in bob._objects
    # TODO: shouldn't we check that alice's object was
    # garbage collected as well?
    # assert x.id not in workers["alice"]._objects


def test_implicit_garbage_collection_pointer(workers):
    """Tests whether GCing a PointerTensor GCs the remote object too."""
    bob = workers["bob"]

    alice, bob = workers["alice"], workers["bob"]

    # create tensor
    x = torch.Tensor([1, 2])

    # send tensor to bob
    x_ptr = x.send(bob)

    # ensure bob has tensor
    assert x.id in bob._objects

    # delete pointer to tensor, which should
    # automatically garbage collect the remote
    # object on Bob's machine
    x_ptr = "asdf"

    # ensure bob's object was garbage collected
    assert x.id not in bob._objects


def test_implicit_garbage_collect_double_pointer(workers):
    """Tests whether GCing a pointer to a pointer garbage collects
    the remote object too"""

    alice, bob = workers["alice"], workers["bob"]

    # create tensor
    x = torch.Tensor([1, 2])

    # send tensor to bob and then pointer to alice
    x_ptr = x.send(bob)
    x_ptr_ptr = x_ptr.send(alice)

    # ensure bob has tensor
    assert x.id in bob._objects

    # delete pointer to pointer to tensor, which should automatically
    # garbage collect the remote object on Bob's machine
    x_ptr_ptr = "asdf"

    # ensure bob's object was garbage collected
    assert x.id not in bob._objects
    # TODO: shouldn't we check that alice's object was
    # garbage collected as well?
    # assert x.id not in alice._objects

    # Chained version
    x = torch.Tensor([1, 2])
    x_id = x.id
    # send tensor to bob and then pointer to alice
    x = x.send(bob).send(alice)
    # ensure bob has tensor
    assert x_id in bob._objects
    # delete pointer to pointer to tensor
    x = "asdf"
    # ensure bob's object was garbage collected
    assert x_id not in bob._objects
    # TODO: shouldn't we check that alice's object was
    # garbage collected as well?
    # assert x.id not in alice._objects


# TESTING IN PLACE METHODS


def test_inplace_method_on_pointer(workers):
    bob = workers["bob"]

    tensor = torch.tensor([[1.0, 2], [4.0, 2]])
    pointer = tensor.send(bob)
    pointer.add_(pointer)
    tensor_back = pointer.get()
    assert (tensor * 2 == tensor_back).all()


# TESTING LOGGING TENSORS


def test_explicit_garbage_collect_logging_on_pointer(workers):
    """
    Tests whether deleting a LoggingTensor on a PointerTensor
    garbage collects the remote object too
    """
    bob = workers["bob"]

    x = torch.Tensor([1, 2])
    x_id = x.id

    x = x.send(bob)
    x = LoggingTensor().on(x)
    assert x_id in bob._objects

    del x

    assert x_id not in bob._objects


def test_implicit_garbage_collect_logging_on_pointer(workers):
    """
    Tests whether GCing a LoggingTensor on a PointerTensor
    garbage collects the remote object too
    """
    bob = workers["bob"]

    x = torch.Tensor([1, 2])
    x_id = x.id

    x = x.send(bob)
    x = LoggingTensor().on(x)
    assert x_id in bob._objects

    x = "open-source"
    assert x_id not in bob._objects


def test_websocket_garbage_collection(hook, start_proc):
    # Args for initializing the websocket server and client
    base_kwargs = {"id": "ws_gc", "host": "localhost", "port": 8555, "hook": hook}
    server = start_proc(WebsocketServerWorker, **base_kwargs)

    time.sleep(0.1)
    client_worker = WebsocketClientWorker(**base_kwargs)

    sample_data = torch.tensor([1, 2, 3, 4])
    sample_ptr = sample_data.send(client_worker)

    _ = sample_ptr.get()
    assert sample_data not in client_worker._objects

    client_worker.ws.shutdown()
    client_worker.ws.close()
    time.sleep(0.1)
    server.terminate()



import torch
import torch as th
import syft

from syft.frameworks.torch.crypto.securenn import (
    private_compare,
    decompose,
    share_convert,
    relu_deriv,
    division,
    maxpool,
    maxpool_deriv,
)


def test_xor_implementation(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    r = decompose(th.tensor([3])).send(alice, bob).child
    x_bit_sh = decompose(th.tensor([23])).share(alice, bob, crypto_provider=james).child
    j0 = torch.zeros(x_bit_sh.shape).long().send(bob)
    j1 = torch.ones(x_bit_sh.shape).long().send(alice)
    j = syft.MultiPointerTensor(children=[j0.child, j1.child])
    w = (j * r) + x_bit_sh - (2 * x_bit_sh * r)

    r_real = r.virtual_get()[0]
    x_real = x_bit_sh.virtual_get()
    w_real = r_real + x_real - 2 * r_real * x_real
    assert (w.virtual_get() == w_real).all()


def test_private_compare(workers):
    """
    Test private compare which returns: β′ = β ⊕ (x > r).
    """
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x_bit_sh = decompose(torch.LongTensor([13])).share(alice, bob, crypto_provider=james).child
    r = torch.LongTensor([12]).send(alice, bob).child

    beta = torch.LongTensor([1]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert not beta_p

    beta = torch.LongTensor([0]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert beta_p

    # Big values
    x_bit_sh = decompose(torch.LongTensor([2 ** 60])).share(alice, bob, crypto_provider=james).child
    r = torch.LongTensor([2 ** 61]).send(alice, bob).child

    beta = torch.LongTensor([1]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert beta_p

    beta = torch.LongTensor([0]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert not beta_p

    # Multidimensional tensors
    x_bit_sh = (
        decompose(torch.LongTensor([[13, 44], [1, 28]]))
        .share(alice, bob, crypto_provider=james)
        .child
    )
    r = torch.LongTensor([[12, 44], [12, 33]]).send(alice, bob).child

    beta = torch.LongTensor([1]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert (beta_p == torch.tensor([[0, 1], [1, 1]])).all()

    beta = torch.LongTensor([0]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert (beta_p == torch.tensor([[1, 0], [0, 0]])).all()

    # Negative values
    x_val = -105 % 2 ** 62
    r_val = -52 % 2 ** 62  # The protocol works only for values in Zq
    x_bit_sh = decompose(torch.LongTensor([x_val])).share(alice, bob, crypto_provider=james).child
    r = torch.LongTensor([r_val]).send(alice, bob).child

    beta = torch.LongTensor([1]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert beta_p

    beta = torch.LongTensor([0]).send(alice, bob).child
    beta_p = private_compare(x_bit_sh, r, beta)
    assert not beta_p


def test_share_convert(workers):
    """
    This is a light test as share_convert is not used for the moment
    """
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    L = 2 ** 62
    x_bit_sh = (
        torch.LongTensor([13, 3567, 2 ** 60])
        .share(alice, bob, crypto_provider=james, field=L)
        .child
    )

    res = share_convert(x_bit_sh)
    assert res.field == L - 1
    assert (res.get() % L == torch.LongTensor([13, 3567, 2 ** 60])).all()


def test_relu_deriv(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    x = th.tensor([10, 0, -3]).share(alice, bob, crypto_provider=james).child
    r = relu_deriv(x)

    assert (r.get() == th.tensor([1, 1, 0])).all()


def test_relu(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    x = th.tensor([1, -3]).share(alice, bob, crypto_provider=james)
    r = x.relu()

    assert (r.get() == th.tensor([1, 0])).all()

    x = th.tensor([1.0, 3.1, -2.1]).fix_prec().share(alice, bob, crypto_provider=james)
    r = x.relu()

    assert (r.get().float_prec() == th.tensor([1, 3.1, 0])).all()


def test_division(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x0 = th.tensor(10).share(alice, bob, crypto_provider=james).child
    y0 = th.tensor(2).share(alice, bob, crypto_provider=james).child
    res0 = division(x0, y0, bit_len_max=5)

    x1 = th.tensor([[25, 9], [10, 30]]).share(alice, bob, crypto_provider=james).child
    y1 = th.tensor([[5, 12], [2, 7]]).share(alice, bob, crypto_provider=james).child
    res1 = division(x1, y1, bit_len_max=5)

    assert res0.get() == torch.tensor(5)
    assert (res1.get() == torch.tensor([[5, 0], [5, 4]])).all()


def test_maxpool(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    x = th.tensor([[10, 0], [15, 7]]).share(alice, bob, crypto_provider=james).child
    max, ind = maxpool(x)

    assert max.get() == torch.tensor(15)
    assert ind.get() == torch.tensor(2)


def test_maxpool_deriv(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    x = th.tensor([[10, 0], [15, 7]]).share(alice, bob, crypto_provider=james).child
    max_d = maxpool_deriv(x)

    assert (max_d.get() == torch.tensor([[0, 0], [1, 0]])).all()
